"""tests for ``vak.io.spect`` module"""
from pathlib import Path

import pandas as pd
import pytest

import vak.io.spect
import vak.files.spect


def expected_spect_paths_in_dataframe(
    vak_df, expected_spect_paths, not_expected_spect_paths=None
):
    """tests that a dataframe ``vak_df`` contains
    all paths in ``expected_spect_paths``, and only those paths,
    in its ``spect_path`` column.
    If so, returns True.

    Parameters
    ----------
    vak_df : pandas.Dataframe
        created by vak.io.spect.to_dataframe
    expected_spect_paths : list
        of paths to spectrogram files, that **should** be in vak_df.spect_path column
    not_expected_spect_paths : list
        of paths to spectrogram files, that should **not** be in vak_df.spect_path column
    """
    assert type(vak_df) == pd.DataFrame

    spect_paths_from_df = [Path(spect_path) for spect_path in vak_df["spect_path"]]

    for expected_spect_path in list(expected_spect_paths):
        assert expected_spect_path in spect_paths_from_df
        spect_paths_from_df.remove(expected_spect_path)

    # test that **only** expected paths were in DataFrame
    if not_expected_spect_paths is not None:
        for not_expected_spect_path in not_expected_spect_paths:
            assert not_expected_spect_path not in spect_paths_from_df

    # test that **only** expected paths were in DataFrame
    # spect_paths_from_df should be empty after popping off all the expected paths
    assert (
        len(spect_paths_from_df) == 0
    )  # yes I know this isn't "Pythonic". It's readable, go away.

    return True  # all asserts passed


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_dir(
    spect_format,
    annot_format,
    specific_spect_dir,
    specific_spect_list,
    specific_annot_list,
    specific_labelset,
):
    """test that ``vak.io.spect.to_dataframe`` works
    when we point it at directory + give it list of annotations"""
    spect_dir = specific_spect_dir(spect_format)
    labelset = specific_labelset(annot_format)
    annot_list = specific_annot_list(annot_format)

    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format,
        spect_dir=spect_dir,
        labelset=labelset,
        annot_list=annot_list,
        annot_format=annot_format,
    )

    spect_list_all_labels_in_labelset = specific_spect_list(
        spect_format, "all_labels_in_labelset"
    )
    spect_list_labels_not_in_labelset = specific_spect_list(
        spect_format, "labels_not_in_labelset"
    )
    assert expected_spect_paths_in_dataframe(
        vak_df, spect_list_all_labels_in_labelset, spect_list_labels_not_in_labelset
    )


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_dir_no_labelset(
    spect_format,
    annot_format,
    specific_spect_dir,
    specific_spect_list,
    specific_annot_list,
):
    """test that ``vak.io.spect.to_dataframe`` works when we point it at directory + give it list of annotations
    but do not give it a labelset to filter out files"""
    spect_dir = specific_spect_dir(spect_format)
    annot_list = specific_annot_list(annot_format)

    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format,
        spect_dir=spect_dir,
        labelset=None,
        annot_list=annot_list,
        annot_format="yarden",
    )

    spect_list = specific_spect_list(spect_format)
    assert expected_spect_paths_in_dataframe(vak_df, spect_list)


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_dir_without_annot(
    spect_format, annot_format, specific_spect_dir, specific_spect_list
):
    """test ``vak.io.spect.to_dataframe`` works with a dataset from spectrogram files without annotations,
    # e.g. if we're going to predict the annotations using the spectrograms"""
    spect_dir = specific_spect_dir(spect_format)

    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format, spect_dir=spect_dir, annot_list=None
    )

    spect_list = specific_spect_list(spect_format)
    assert expected_spect_paths_in_dataframe(vak_df, spect_list)


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_files(
    spect_format,
    annot_format,
    specific_spect_list,
    specific_annot_list,
    specific_labelset,
):
    """test that ``vak.io.spect.to_dataframe`` works
    when we give it list of spectrogram files and a list of annotations"""
    spect_list = specific_spect_list(spect_format)
    labelset = specific_labelset(annot_format)
    annot_list = specific_annot_list(annot_format)

    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format,
        spect_files=spect_list,
        labelset=labelset,
        annot_list=annot_list,
        annot_format=annot_format,
    )

    spect_list_all_labels_in_labelset = specific_spect_list(
        spect_format, "all_labels_in_labelset"
    )
    spect_list_labels_not_in_labelset = specific_spect_list(
        spect_format, "labels_not_in_labelset"
    )
    assert expected_spect_paths_in_dataframe(
        vak_df, spect_list_all_labels_in_labelset, spect_list_labels_not_in_labelset
    )


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_files_no_labelset(
    spect_format, annot_format, specific_spect_list, specific_annot_list
):
    """test that ``vak.io.spect.to_dataframe`` works
    when we give it list of spectrogram files and a list of annotations
    but do not give it a labelset to filter out files"""
    spect_list = specific_spect_list(spect_format)
    annot_list = specific_annot_list(annot_format)

    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format,
        spect_files=spect_list,
        labelset=None,
        annot_list=annot_list,
        annot_format=annot_format,
    )

    spect_list = specific_spect_list(spect_format)
    assert expected_spect_paths_in_dataframe(vak_df, spect_list)


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_annot_map(
    spect_format,
    annot_format,
    specific_spect_list,
    specific_annot_list,
    specific_labelset,
):
    """test that ``vak.io.spect.to_dataframe`` works
    when we give it a dict that maps spectrogram files to annotations
    but do not give it a labelset to filter out files"""
    spect_list = specific_spect_list(spect_format)
    labelset = specific_labelset(annot_format)
    annot_list = specific_annot_list(annot_format)

    spect_annot_map = dict(zip(spect_list, annot_list))
    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format,
        labelset=labelset,
        spect_annot_map=spect_annot_map,
        annot_format=annot_format,
    )

    spect_list_all_labels_in_labelset = specific_spect_list(
        spect_format, "all_labels_in_labelset"
    )
    spect_list_labels_not_in_labelset = specific_spect_list(
        spect_format, "labels_not_in_labelset"
    )
    assert expected_spect_paths_in_dataframe(
        vak_df, spect_list_all_labels_in_labelset, spect_list_labels_not_in_labelset
    )


@pytest.mark.parametrize(
    "spect_format, annot_format",
    [
        ("mat", "yarden"),
        ("npz", "notmat"),
    ],
)
def test_to_dataframe_spect_annot_map_no_labelset(
    spect_format, annot_format, specific_spect_list, specific_annot_list
):
    """test that ``vak.io.spect.to_dataframe`` works
    when we give it a dict that maps spectrogram files to annotations
    but do not give it a labelset to filter out files"""
    spect_list = specific_spect_list(spect_format)
    annot_list = specific_annot_list(annot_format)

    spect_annot_map = dict(zip(spect_list, annot_list))

    vak_df = vak.io.spect.to_dataframe(
        spect_format=spect_format,
        labelset=None,
        spect_annot_map=spect_annot_map,
        annot_format=annot_format,
    )

    spect_list = specific_spect_list(spect_format)
    assert expected_spect_paths_in_dataframe(vak_df, spect_list)


def test_to_dataframe_no_spect_dir_files_or_map_raises(annot_list_yarden):
    """test that calling ``to_dataframe`` without one of:
    spect dir, spect files, or spect files/annotations mapping
    raises ValueError"""
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat",
            spect_dir=None,
            spect_files=None,
            annot_list=annot_list_yarden,
            spect_annot_map=None,
            annot_format="yarden",
        )


def test_to_dataframe_invalid_spect_format_raises(spect_dir_mat, annot_list_yarden):
    """test that calling ``to_dataframe`` with an invalid spect format raises a ValueError"""
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="npy",  # 'npy' not a valid spect format
            spect_dir=spect_dir_mat,
            annot_list=annot_list_yarden,
            annot_format="yarden",
        )


def test_to_dataframe_dir_and_list_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``to_dataframe`` with both dir and list raises a ValueError"""
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat",
            spect_dir=spect_dir_mat,
            spect_files=spect_list_mat,
            annot_list=annot_list_yarden,
            annot_format="yarden",
        )


def test_to_dataframe_dir_and_map_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``to_dataframe`` with both dir and map raises a ValueError"""
    spect_annot_map = dict(zip(spect_list_mat, annot_list_yarden))
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat",
            spect_dir=spect_dir_mat,
            spect_annot_map=spect_annot_map,
            annot_format="yarden",
        )


def test_to_dataframe_list_and_map_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``to_dataframe`` with both list and map raises a ValueError"""
    spect_annot_map = dict(zip(spect_list_mat, annot_list_yarden))
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat",
            spect_files=spect_list_mat,
            spect_annot_map=spect_annot_map,
            annot_format="yarden",
        )


def test_to_dataframe_annot_list_and_map_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``to_dataframe`` with both list of annotations and map raises a ValueError"""
    spect_annot_map = dict(zip(spect_list_mat, annot_list_yarden))
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat",
            spect_annot_map=spect_annot_map,
            annot_list=annot_list_yarden,
            annot_format="yarden",
        )


def test_to_dataframe_annot_list_without_annot_format_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``to_dataframe`` with a list of annotations but no annot_format raises a ValueError"""
    spect_annot_map = dict(zip(spect_list_mat, annot_list_yarden))
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat", annot_list=annot_list_yarden, annot_format=None
        )


def test_to_dataframe_spect_annot_map_without_annot_format_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``to_dataframe`` with a list of annotations but no annot_format raises a ValueError"""
    spect_annot_map = dict(zip(spect_list_mat, annot_list_yarden))
    with pytest.raises(ValueError):
        vak.io.spect.to_dataframe(
            spect_format="mat", spect_annot_map=spect_annot_map, annot_format=None
        )



from pathlib import Path

import pytest

HERE = Path(__file__).parent


@pytest.fixture
def test_data_root():
    """Path that points to root of data_for_tests directory"""
    return HERE.joinpath("..", "data_for_tests")


@pytest.fixture
def source_test_data_root(test_data_root):
    """'source' test data, i.e., files **not** created by vak, that is,
    the input data used when vak does create files (csv files, logs,
    neural network checkpoints, etc.)
    """
    return test_data_root.joinpath("source")


@pytest.fixture
def generated_test_data_root(test_data_root):
    """'generated' test data, i.e. files created by vak:
    csv files, logs, neural network checkpoints, etc."""
    return test_data_root.joinpath("generated")


@pytest.fixture
def generated_prep_data_root(generated_test_data_root):
    return generated_test_data_root / "prep"


@pytest.fixture
def generated_results_data_root(generated_test_data_root):
    return generated_test_data_root / "results"



from pathlib import Path
from distutils.util import strtobool


def bool_from_str(value):
    if type(value) == bool:
        return value
    elif type(value) == str:
        return bool(strtobool(value))


def comma_separated_list(value):
    if type(value) is list:
        return value
    elif type(value) is str:
        return [element.strip() for element in value.split()]
    else:
        raise TypeError(
            f"unexpected type when converting to comma-separated list: {type(value)}"
        )


def expanded_user_path(value):
    return Path(value).expanduser()


def range_str(range_str, sort=True):
    """Generate range of ints from a formatted string,
    then convert range from int to str

    Examples
    --------
    >>> range_str('1-4,6,9-11')
    ['1','2','3','4','6','9','10','11']

    Takes a range in form of "a-b" and returns
    a list of numbers between a and b inclusive.
    Also accepts comma separated ranges like "a-b,c-d,f"  which will
    return a list with numbers from a to b, c to d, and f.

    Parameters
    ----------
    range_str : str
        of form 'a-b,c', where a hyphen indicates a range
        and a comma separates ranges or single numbers
    sort : bool
        If True, sort output before returning. Default is True.

    Returns
    -------
    list_range : list
        of integer values converted to single-character strings, produced by parsing range_str
    """
    # adapted from
    # http://code.activestate.com/recipes/577279-generate-list-of-numbers-from-hyphenated-and-comma/
    s = "".join(range_str.split())  # removes white space
    list_range = []
    for substr in range_str.split(","):
        subrange = substr.split("-")
        if len(subrange) not in [1, 2]:
            raise SyntaxError(
                "unable to parse range {} in labelset {}.".format(subrange, substr)
            )
        list_range.extend([int(subrange[0])]) if len(
            subrange
        ) == 1 else list_range.extend(range(int(subrange[0]), int(subrange[1]) + 1))

    if sort:
        list_range.sort()

    return [str(list_int) for list_int in list_range]


def labelset_to_set(labelset):
    """convert value for 'labelset' argument into a Python set.
    Used by ``vak`` internally to convert

    Parameters
    ----------
    labelset : str, list
        string or list specifying a unique set of labels
        used to annotate a dataset of vocalizations.
        See Notes for details on valid values.

    Returns
    -------
    labelset : set
        of strings, labels used to annotate segments.

    Notes
    -----
    If ``labelset```` is a str, and it starts with "range:", then everything after range is converted to
    some range of integers, by passing the string to ``vak.config.converters.range_str``,
    and the returned list is converted to a set. E.g. "range: 1-5" becomes {'1', '2', '3', '4', '5'}.
    Other strings that do not start with "range:" are just converted to a set. E.g. "abc" becomes {'a', 'b', 'c'}.

    If ``labelset`` is a list, then all values in the list must strings or integers.
    Any that begin with "range:" will be passed to vak.config.converters.range_str.
    Any other multiple-character strings in a list are **not** split,
    unlike when the value for the ``labelset`` option is just a single string
    with multiple characters.
    If you have segments annotated with multiple characters,
    you should specify them using a list, e.g., ['en', 'ab', 'cd']

    If ``labelset`` is a set, it is returned as is, so that this function does not return ``None``,
    which would cause other functions to behave as if no ``labelset`` were specified.

    Examples
    --------
    >>> labelset_from_toml_value('abc')
    {'a', 'b', 'c'}
    >>> labelset_from_toml_value('1235')
    {'1', '2', '3', '5'}
    >>> labelset_from_toml_value('range: 1-3, 5')
    {'1', '2', '3', '5'}
    >>> labelset_from_toml_value([1, 2, 3, 5])
    {'1', '2', '3'}
    >>> labelset_from_toml_value(['a', 'b', 'c'])
    {'a', 'b', 'c'}
    >>> labelset_from_toml_value(['range: 1-3', 'noise'])
    {'1', '2', '3', 'noise'}
    """
    if type(labelset) not in (str, list, set):
        raise TypeError(
            "labelset must be specified as a string, list, or set, "
            f"but the type of labelset was: {type(labelset)}"
        )

    if type(labelset) is set:
        return labelset
    elif type(labelset) is str:
        if labelset.startswith("range:"):
            labelset = labelset.replace("range:", "")
            return set(range_str(labelset))
        else:
            return set(labelset)
    elif type(labelset) is list:
        labelset_out = []
        for label in labelset:
            if isinstance(label, int):
                labelset_out.append(str(label))
            elif isinstance(label, str):
                if label.startswith("range:"):
                    label = label.replace("range:", "")
                    labelset_out.extend(range_str(label))
                else:
                    labelset_out.append(label)
            else:
                raise TypeError(
                    f"label '{label}' specified in labelset is invalid type: {type(label)}."
                    "Labels must be strings or integers."
                )
        return set(labelset_out)



"""parses [EVAL] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .validators import is_a_directory, is_a_file, is_valid_model_name
from .. import device
from ..converters import comma_separated_list, expanded_user_path


@attr.s
class EvalConfig:
    """class that represents [EVAL] section of config.toml file

    Attributes
    ----------
    csv_path : str
        path to where dataset was saved as a csv.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str
        Path to location where .csv files with evaluation metrics should be saved.
    labelmap_path : str
        path to 'labelmap.json' file.
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    batch_size : int
        number of samples per batch presented to models during training.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    """

    # required, external files
    checkpoint_path = attr.ib(converter=expanded_user_path, validator=is_a_file)
    labelmap_path = attr.ib(converter=expanded_user_path, validator=is_a_file)
    output_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)

    # required, model / dataloader
    models = attr.ib(
        converter=comma_separated_list,
        validator=[instance_of(list), is_valid_model_name],
    )
    batch_size = attr.ib(converter=int, validator=instance_of(int))

    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )

    # optional, transform
    spect_scaler_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )

    # optional, data loader
    num_workers = attr.ib(validator=instance_of(int), default=2)
    device = attr.ib(validator=instance_of(str), default=device.get_default())



import attr
from attr.validators import instance_of, optional

from .dataloader import DataLoaderConfig
from .eval import EvalConfig
from .learncurve import LearncurveConfig
from .predict import PredictConfig
from .prep import PrepConfig
from .spect_params import SpectParamsConfig
from .train import TrainConfig


@attr.s
class Config:
    """class to represent config.toml file

    Attributes
    ----------
    prep : vak.config.prep.PrepConfig
        represents ``[PREP]`` section of config.toml file
    spect_params : vak.config.spect_params.SpectParamsConfig
        represents ``[SPECT_PARAMS]`` section of config.toml file
    dataloader : vak.config.dataloader.DataLoaderConfig
        represents ``[DATALOADER]`` section of config.toml file
    train : vak.config.train.TrainConfig
        represents ``[TRAIN]`` section of config.toml file
    eval : vak.config.eval.EvalConfig
        represents ``[EVAL]`` section of config.toml file
    predict : vak.config.predict.PredictConfig
        represents ``[PREDICT]`` section of config.toml file.
    learncurve : vak.config.learncurve.LearncurveConfig
        represents ``[LEARNCURVE]`` section of config.toml file
    """

    spect_params = attr.ib(
        validator=instance_of(SpectParamsConfig), default=SpectParamsConfig()
    )
    dataloader = attr.ib(
        validator=instance_of(DataLoaderConfig), default=DataLoaderConfig()
    )

    prep = attr.ib(validator=optional(instance_of(PrepConfig)), default=None)
    train = attr.ib(validator=optional(instance_of(TrainConfig)), default=None)
    eval = attr.ib(validator=optional(instance_of(EvalConfig)), default=None)
    predict = attr.ib(validator=optional(instance_of(PredictConfig)), default=None)
    learncurve = attr.ib(
        validator=optional(instance_of(LearncurveConfig)), default=None
    )



from pathlib import Path

import toml

from .. import models
from ..engine.model import Model


def map_from_config_dict(config_dict, model_names):
    """map a list of model names to model configuration sections from a
    config.toml file.

    Given the configuraiton in a dict and a list of model names, returns dict that
    maps model names to config sections. If no section in the config.toml file
    matches the model name, an error is raised.

    Used to get configuration for only the models specified
    in a certain section of config.toml file, e.g. in the TRAIN section.

    The returned model-config map can be used with vak.models.from_model_config_map
    (and the number of classes and input shape) to get a list of model instances
    ready for training.

    Parameters
    ----------
    config_dict : dict
        configuration from a .toml file, loaded into a dictionary
    model_names : list
        of str, i.e. names of models specified by a section
        (such as TRAIN or PREDICT) that should each have corresponding sections
        specifying their configuration: hyperparameters such as learning rate, etc.

    Returns
    -------
    model_config_map : dict
        where each key is the name of a model and the corresponding value is
        a section from a config.toml file.
    """
    # first check whether models in list are installed

    # load entry points within function, not at module level,
    # to avoid circular dependencies
    # (user would be unable to import models in other packages
    # if the module in the other package needed to `import vak`)
    MODELS = {model_name: model_builder for model_name, model_builder in models.find()}
    for model_name in model_names:
        if model_name not in MODELS:
            # try appending 'Model' to name
            tmp_model_name = f"{model_name}Model"
            if tmp_model_name not in MODELS:
                raise ValueError(
                    f"Did not find an installed model named {model_name} or {tmp_model_name}. "
                    f"Installed models are: {list(MODELS.keys())}"
                )

    # now see if we can find corresponding sections in config
    sections = list(config_dict.keys())
    model_config_map = {}
    for model_name in model_names:
        if model_name in sections:
            model_config_dict = config_dict[model_name]
        else:
            # try appending 'Model' to name
            tmp_model_name = f"{model_name}Model"
            if tmp_model_name not in sections:
                raise ValueError(
                    f"did not find section named {model_name} or {tmp_model_name} "
                    f"in config"
                )
            model_config_dict = config_dict[tmp_model_name]

        # check if config declares parameters for required attributes;
        # if not, just put an empty dict that will get passed as the "kwargs"
        for attr in Model.REQUIRED_SUBCLASS_ATTRIBUTES:
            if attr not in model_config_dict:
                model_config_dict[attr] = {}

        model_config_map[model_name] = config_dict[model_name]

    return model_config_map


def map_from_path(toml_path, model_names):
    """map a list of model names to sections from a .toml configuration file
     that specify parameters for those models.

    Parameters
    ----------
    toml_path : str, Path
        to configuration file in .toml format
     model_names : list
        of str, i.e. names of models specified by a section
        (such as TRAIN or PREDICT) that should each have corresponding sections
        specifying their configuration: hyperparameters such as learning rate, etc.

    Returns
    -------
    model_config_map : dict
        where each key is the name of a model and the corresponding value is
        a section from a config.toml file.
    """
    # check config_path is a file,
    # because if it doesn't ConfigParser will just return an "empty" instance w/out sections or options
    toml_path = Path(toml_path)
    if not toml_path.is_file():
        raise FileNotFoundError(
            f"file not found, or not recognized as a file: {toml_path}"
        )

    with toml_path.open("r") as fp:
        config_dict = toml.load(fp)
    return map_from_config_dict(config_dict, model_names)



"""
Invokes __main__ when the module is run as a script.
Example: python -m vak --help
"""
import argparse
from pathlib import Path


def get_parser():
    """returns ArgumentParser instance used by main()"""
    from .cli import cli  # avoid circular import

    parser = argparse.ArgumentParser(
        description="vak command-line interface",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "command",
        type=str,
        metavar="command",
        choices=cli.CLI_COMMANDS,
        help="Command to run, valid options are:\n"
        f"{cli.CLI_COMMANDS}\n"
        "$ vak train ./configs/config_2018-12-17.toml",
    )
    parser.add_argument(
        "configfile",
        type=Path,
        help="name of config.toml file to use \n"
        "$ vak train ./configs/config_2018-12-17.toml",
    )
    return parser


def main():
    """main function, called when package is run with `python -m vak` or
    alternatively just calling `vak` at the command line (because this
    function is installed under just `vak` as a console script entry point,
    see setup.py)
    """
    from .cli import cli  # avoid circular import

    parser = get_parser()
    args = parser.parse_args()
    cli.cli(command=args.command, config_file=args.configfile)


if __name__ == "__main__":
    main()



"""parses [SPECT_PARAMS] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of


def freq_cutoffs_validator(instance, attribute, value):
    if len(value) != 2:
        raise ValueError(
            f"freq_cutoffs should be a list of two elements, but instead got: {value}"
        )
    if value[0] > value[1]:
        raise ValueError(
            f"lower freq_cutoff should be less than higher freq_cutoff, instead of: {value}"
        )


VALID_TRANSFORM_TYPES = {"log_spect", "log_spect_plus_one"}


def is_valid_transform_type(instance, attribute, value):
    if value not in VALID_TRANSFORM_TYPES:
        raise ValueError(
            f"Value for `transform_type`, {value}, in [SPECT_PARAMS] "
            "section of .toml file is not recognized. Must be one "
            f"of the following: {VALID_TRANSFORM_TYPES}"
        )


@attr.s
class SpectParamsConfig:
    """represents parameters for making spectrograms from audio and saving in files

    Attributes
    ----------
    fft_size : int
        size of window for Fast Fourier transform, number of time bins. Default is 512.
    step_size : int
        step size for Fast Fourier transform. Default is 64.
    freq_cutoffs : tuple
        of two elements, lower and higher frequencies. Used to bandpass filter audio
        (using a Butter filter) before generating spectrogram.
        Default is None, in which case no bandpass filtering is applied.
    transform_type : str
        one of {'log_spect', 'log_spect_plus_one'}.
        'log_spect' transforms the spectrogram to log(spectrogram), and
        'log_spect_plus_one' does the same thing but adds one to each element.
        Default is None. If None, no transform is applied.
    thresh: int
        threshold minimum power for log spectrogram.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    freqbins_key : str
        key for accessing vector of frequency bins in files. Default is 'f'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    audio_path_key : str
        key for accessing path to source audio file for spectogram in files.
        Default is 'audio_path'.
    """

    fft_size = attr.ib(converter=int, validator=instance_of(int), default=512)
    step_size = attr.ib(converter=int, validator=instance_of(int), default=64)
    freq_cutoffs = attr.ib(
        validator=validators.optional(freq_cutoffs_validator), default=None
    )
    thresh = attr.ib(
        converter=converters.optional(float),
        validator=validators.optional(instance_of(float)),
        default=None,
    )
    transform_type = attr.ib(
        validator=validators.optional([instance_of(str), is_valid_transform_type]),
        default=None,
    )
    spect_key = attr.ib(validator=instance_of(str), default="s")
    freqbins_key = attr.ib(validator=instance_of(str), default="f")
    timebins_key = attr.ib(validator=instance_of(str), default="t")
    audio_path_key = attr.ib(validator=instance_of(str), default="audio_path")



"""command-line interface functions for training,
creating learning curves, etc."""

from . import cli, eval, learncurve, predict, prep, train


__all__ = [
    "cli",
    "eval",
    "learncurve",
    "predict",
    "prep",
    "train",
]



"""tests for vak.core.prep module"""
from pathlib import Path
import shutil

import pandas as pd
from pandas.testing import assert_series_equal
import pytest

import vak.config
import vak.constants
import vak.core.train
import vak.paths
import vak.io.spect


# written as separate function so we can re-use in tests/unit/test_cli/test_prep.py
def prep_output_matches_expected(csv_path, df_returned_by_prep):
    assert Path(csv_path).exists()
    df_from_csv_path = pd.read_csv(csv_path)

    for column in vak.io.spect.DF_COLUMNS:
        if column == "duration":
            check_exact = False
        else:
            check_exact = True
        assert_series_equal(
            df_from_csv_path[column],
            df_returned_by_prep[column],
            check_exact=check_exact,
        )

    return True


@pytest.mark.parametrize(
    "config_type, audio_format, spect_format, annot_format",
    [
        ("eval", "cbin", None, "notmat"),
        ("learncurve", "cbin", None, "notmat"),
        ("predict", "cbin", None, "notmat"),
        ("predict", "wav", None, "birdsong-recognition-dataset"),
        ("train", "cbin", None, "notmat"),
        ("train", "wav", None, "birdsong-recognition-dataset"),
        ("train", None, "mat", "yarden"),
    ],
)
def test_prep(
    config_type,
    audio_format,
    spect_format,
    annot_format,
    specific_config,
    default_model,
    tmp_path,
):
    output_dir = tmp_path.joinpath(
        f"test_prep_{config_type}_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = {
        "section": "PREP",
        "option": "output_dir",
        "value": str(output_dir),
    }
    toml_path = specific_config(
        config_type=config_type,
        model=default_model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)

    purpose = config_type.lower()
    vak_df, csv_path = vak.core.prep(
        data_dir=cfg.prep.data_dir,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_output_dir=cfg.prep.spect_output_dir,
        spect_params=cfg.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
        logger=None,
    )

    assert prep_output_matches_expected(csv_path, vak_df)


def test_prep_with_single_audio_and_annot(source_test_data_root,
                                          specific_config,
                                          default_model,
                                          tmp_path):
    """
    regression test, checks that we avoid a repeat of
    https://github.com/NickleDave/vak/issues/467
    """
    data_dir = tmp_path / 'data_dir_with_single_audio_and_annot'
    data_dir.mkdir()
    source_data_dir = source_test_data_root / 'audio_cbin_annot_notmat/gy6or6/032412'
    cbins = sorted(source_data_dir.glob('*.cbin'))
    a_cbin = cbins[0]
    shutil.copy(a_cbin, data_dir)
    a_rec = a_notmat = a_cbin.parent / (a_cbin.stem + '.rec')
    assert a_rec.exists()
    shutil.copy(a_rec, data_dir)
    a_notmat = a_cbin.parent / (a_cbin.name + '.not.mat')
    assert a_notmat.exists()
    shutil.copy(a_notmat, data_dir)

    output_dir = tmp_path.joinpath(
        f"test_prep_eval_single_audio_and_annot"
    )
    output_dir.mkdir()

    options_to_change = [
        {
            "section": "PREP",
            "option": "data_dir",
            "value": str(data_dir),
        },
        {
            "section": "PREP",
            "option": "output_dir",
            "value": str(output_dir),
        },
    ]

    toml_path = specific_config(
        config_type='eval',
        model=default_model,
        audio_format='cbin',
        annot_format='notmat',
        spect_format=None,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)

    purpose = 'eval'
    vak_df, csv_path = vak.core.prep(
        data_dir=cfg.prep.data_dir,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_output_dir=cfg.prep.spect_output_dir,
        spect_params=cfg.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
        logger=None,
    )

    assert len(vak_df) == 1


def test_prep_when_annot_has_single_segment(source_test_data_root,
                                            specific_config,
                                            default_model,
                                            tmp_path):
    """
    regression test, checks that we avoid a repeat of
    https://github.com/NickleDave/vak/issues/466
    """
    data_dir = source_test_data_root / 'audio_cbin_annot_notmat' / 'gy6or6-song-edited-to-have-single-segment'

    output_dir = tmp_path.joinpath(
        f"test_prep_eval_annot_with_single_segment"
    )
    output_dir.mkdir()

    options_to_change = [
        {
            "section": "PREP",
            "option": "data_dir",
            "value": str(data_dir),
        },
        {
            "section": "PREP",
            "option": "output_dir",
            "value": str(output_dir),
        },
    ]

    toml_path = specific_config(
        config_type='eval',
        model=default_model,
        audio_format='cbin',
        annot_format='notmat',
        spect_format=None,
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)

    purpose = 'eval'
    vak_df, csv_path = vak.core.prep(
        data_dir=cfg.prep.data_dir,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_output_dir=cfg.prep.spect_output_dir,
        spect_params=cfg.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
        logger=None,
    )

    assert len(vak_df) == 1



from .eval import eval
from .train import train
from .learncurve import learning_curve
from .predict import predict
from .prep import prep


COMMAND_FUNCTION_MAP = {
    "prep": prep,
    "train": train,
    "eval": eval,
    "predict": predict,
    "learncurve": learning_curve,
}

CLI_COMMANDS = tuple(COMMAND_FUNCTION_MAP.keys())


def cli(command, config_file):
    """command-line interface

    Parameters
    ----------
    command : string
        One of {'prep', 'train', 'eval', 'predict', 'learncurve'}
    config_file : str, Path
        path to a config.toml file
    """
    if command in COMMAND_FUNCTION_MAP:
        COMMAND_FUNCTION_MAP[command](toml_path=config_file)
    else:
        raise ValueError(f"command not recognized: {command}")



"""functions for working with paths"""
from pathlib import Path

from . import constants, timenow


def generate_results_dir_name_as_path(root_results_dir=None):
    """generates a name for a new results directory,
    returns as a path

    Parameters
    ----------
    root_results_dir : str, pathlib.Path
        root directory within which a new results directory will be made.
        Default is None, in which case the new results directory name
        will be relative to the current working directory.

    Returns
    -------
    results_path : pathlib.Path
        path to a new results directory, with name of the following format:
        ``f'{vak.constants.RESULTS_DIR_PREFIX}{vak.timenow.get_timenow_as_str}'``.
        e.g., ``results_210211_142329``.
        This function simply builds the name and path with a consistent format.
        To actually make this directory, call ``results_path.mkdir()``
    """
    if root_results_dir:
        root_results_dir = Path(root_results_dir)
    else:
        root_results_dir = Path(".")
    if not root_results_dir.is_dir():
        raise NotADirectoryError(
            f"root_results_dir not recognized as a directory: {root_results_dir}"
        )

    results_dirname = f"{constants.RESULTS_DIR_PREFIX}{timenow.get_timenow_as_str()}"
    return root_results_dir.joinpath(results_dirname)



def _iter(group_name):
    """helper function that iterates through entry points, given name of entry point group"""
    try:
        import pkg_resources
    except (ImportError, IOError):
        return []

    return pkg_resources.iter_entry_points(group_name)



from collections import Counter
import copy
import os
from pathlib import Path

import crowsetta
import numpy as np

from . import files
from . import constants


def format_from_df(vak_df):
    """determine annotation format of a Vak DataFrame.
    Returns string name of annotation format.
    If no annotation format is specified, returns None.
    Raises an error if there are multiple formats.

    Parameters
    ----------
    vak_df : DataFrame
        representating a dataset of vocalizations, with column 'annot_format'.

    Returns
    -------
    annot_format : str
        format of annotations for vocalizations.
    """
    annot_format = vak_df["annot_format"].unique()
    if len(annot_format) == 1:
        annot_format = annot_format.item()
        if annot_format is None or annot_format == constants.NO_ANNOTATION_FORMAT:
            return None
    elif len(annot_format) > 1:
        raise ValueError(
            f"unable to load labels for dataset, found multiple annotation formats: {annot_format}"
        )

    return annot_format


def from_df(vak_df):
    """get list of annotations from a vak DataFrame.
    If no annotation format is specified for the DataFrame
    (in the 'annot_format' column), returns None.

    Parameters
    ----------
    vak_df : DataFrame
        representating a dataset of vocalizations, with column 'annot_format'.

    Returns
    -------
    annots : list
        of annotations for each row in the dataframe,
        represented as crowsetta.Annotation instances.

    Notes
    -----
    This function encapsulates logic for handling different types of
    annotations; it determines whether each row has a separate annotation file,
    or if instead there is a single annotation file associated with all rows.
    If the latter, then the function opens that file and makes sure that
    each row from the dataframe can be paired with an annotation (using `source_annot_map`).
    """
    annot_format = format_from_df(vak_df)
    if annot_format is None:
        return None

    scribe = crowsetta.Transcriber(format=annot_format)

    if len(vak_df["annot_path"].unique()) == 1:
        # --> there is a single annotation file associated with all rows
        # this can be true in two different cases:
        # (1) many rows, all have the same file
        # (2) only one row, so there's only one annotation file (which may contain annotation for multiple source files)
        annot_path = vak_df["annot_path"].unique().item()
        annots = scribe.from_file(annot_path)

        # as long as we have at least as many annotations as there are rows in the dataframe
        if (isinstance(annots, list) and len(annots) >= len(vak_df)) or (  # case 1
            isinstance(annots, crowsetta.Annotation) and len(vak_df) == 1
        ):  # case 2
            if isinstance(annots, crowsetta.Annotation):
                annots = [
                    annots
                ]  # wrap in list for source_annot_map to iterate over it
            # then we can try and map those annotations to the rows
            audio_annot_map = source_annot_map(vak_df["audio_path"].values, annots)
            # sort by row of dataframe
            annots = [
                audio_annot_map[audio_path]
                for audio_path in vak_df["audio_path"].values
            ]

        else:
            raise ValueError(
                "unable to load labels from dataframe; found a single annotation file associated with all "
                "rows in dataframe, but loading it did not return a list of annotations for each row.\n"
                f"Single annotation file: {annot_path}\n"
                f"Loading it returned a {type(annots)}."
            )

    elif len(vak_df["annot_path"].unique()) == len(vak_df):
        # --> there is a unique annotation file (path) for each row, iterate over them to get labels from each
        annots = [
            scribe.from_file(annot_path) for annot_path in vak_df["annot_path"].values
        ]

    else:
        raise ValueError(
            "unable to load labels from dataframe; did not find an annotation file for each row or "
            "a single annotation file associated with all rows."
        )

    return annots


def files_from_dir(annot_dir, annot_format):
    """get all annotation files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.
    """
    if annot_format not in constants.VALID_ANNOT_FORMATS:
        raise ValueError(
            f"specified annotation format, {annot_format} not valid.\n"
            f"Valid formats are: {constants.VALID_ANNOT_FORMATS}"
        )

    format_module = getattr(crowsetta.formats, annot_format)
    ext = format_module.meta.ext
    annot_files = files.from_dir(annot_dir, ext)
    return annot_files


def recursive_stem(path):
    """removes extensions from a filename recursively,
    by calling `os.path.splitext`,
    until the extension is an audio file format handled by vak.
    Then return the stem, that is, the part that precedes the extension.
    Used to match audio, spectrogram, and annotation files by their stems.

    Stops after finding audio extensions so that it does not remove "extensions"
    that are actually other parts of a filename, e.g. a time or data separated by periods.

    Examples
    --------
    >>> recursive_stem('gy6or6_baseline_230312_0808.138.cbin.not.mat')
    'gy6or6_baseline_230312_0808.138'
    >>> recursive_stem('Bird0/spectrograms/0.wav.npz')
    '0'
    >>> recursive_stem('Bird0/Wave/0.wav')
    '0'

    Parameters
    ----------
    path : str, Path
        from which stem should be extracted

    Returns
    -------
    stem : str
        filename that precedes audio extension
    """
    name = Path(path).name
    stem, ext = os.path.splitext(name)
    ext = ext.replace(".", "").lower()
    while ext not in constants.VALID_AUDIO_FORMATS:
        new_stem, ext = os.path.splitext(stem)
        ext = ext.replace(".", "").lower()
        if new_stem == stem:
            raise ValueError(f"unable to compute stem of {path}")
        else:
            stem = new_stem
    return stem


def source_annot_map(source_files, annot_list):
    """map source files, i.e. audio or spectrogram files, to annotations

    returns a ``dict`` where each key is a path to a source audio file
    or array file containing a spectrogram, and the corresponding value
    is the annotation for the source file

    Parameters
    ----------
    source_files : list
        of audio or spectrogram files. The names of the files must begin with the
        audio_path attribute of the corresponding annotations. E.g., if an audio file is
        'bird0-2016-05-04-133027.wav', then there must be an annotation whose
        file attribute equals that filename. Spectrogram files should include
        the audio file name, e.g. 'bird0-2016-05-04-133027.wav.mat' or
        'bird0-2016-05-04-133027.spect.npz' would match an annotation with the
        audio_path attribute 'bird0-2016-05-04-133027.wav'.
    annot_list : list
        of Annotations corresponding to files in source_files
    """
    if type(source_files) == np.ndarray:  # e.g., vak DataFrame['spect_path'].values
        source_files = source_files.tolist()

    # to pair audio files with annotations, make list of tuples
    source_annot_map = {}

    # ----> make a dict with audio stems as keys,
    #       so we can look up annotations by stemming source files and using as keys.
    # First check that we don't have duplicate keys that would cause this to fail silently
    keys = [recursive_stem(annot.audio_path) for annot in annot_list]
    keys_set = set(keys)
    if len(keys_set) < len(keys):
        duplicates = [item for item, count in Counter(keys).items() if count > 1]
        raise ValueError(
            f"found multiple annotations with the same audio filename(s): {duplicates}"
        )
    del keys, keys_set
    audio_stem_annot_map = {
        recursive_stem(annot.audio_path): annot for annot in annot_list
    }

    # Make a copy from which we remove source files after mapping them to annotation,
    # to validate that function worked,
    # by making sure there are no items left in this copy after the loop
    source_files_copy = copy.deepcopy(source_files)
    for source_file in list(
        source_files
    ):  # list() to copy, so we can pop off items while iterating
        # remove stem so we can find .spect files that match with audio files,
        # e.g. find 'llb3_0003_2018_04_23_14_18_54.mat' that should match
        # with 'llb3_0003_2018_04_23_14_18_54.wav'
        source_file_stem = recursive_stem(source_file)

        try:
            annot = audio_stem_annot_map[source_file_stem]
        except KeyError:
            raise ValueError(
                f"could not find annotation for source file: {source_file}.\n"
                f"No annotation had an audio file whose stem matched the source file stem: {source_file_stem}"
            )

        source_annot_map[source_file] = annot
        source_files_copy.remove(source_file)

    if len(source_files_copy) > 0:
        raise ValueError(
            "could not map the following source files to annotations: "
            f"{source_files_copy}"
        )

    return source_annot_map



"""functions for making spectrogram

filters adapted from SciPy cookbook
https://scipy-cookbook.readthedocs.io/items/ButterworthBandpass.html
spectrogram adapted from code by Kyle Kastner and Tim Sainburg
https://github.com/timsainb/python_spectrograms_and_inversion
"""
import numpy as np

from scipy.signal import butter, lfilter
from matplotlib.mlab import specgram


def butter_bandpass(lowcut, highcut, fs, order=5):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype="band")
    return b, a


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y


def spectrogram(
    dat,
    samp_freq,
    fft_size=512,
    step_size=64,
    thresh=None,
    transform_type=None,
    freq_cutoffs=None,
):
    """creates a spectrogram

    Parameters
    ----------
    dat : numpy.ndarray
        audio signal
    samp_freq : int
        sampling frequency in Hz
    fft_size : int
        size of window for Fast Fourier transform, number of time bins.
    step_size : int
        step size for Fast Fourier transform
    transform_type : str
        one of {'log_spect', 'log_spect_plus_one'}.
        'log_spect' transforms the spectrogram to log(spectrogram), and
        'log_spect_plus_one' does the same thing but adds one to each element.
        Default is None. If None, no transform is applied.
    thresh: int
        threshold minimum power for log spectrogram
    freq_cutoffs : tuple
        of two elements, lower and higher frequencies.

    Return
    ------
    spect : numpy.ndarray
        spectrogram
    freqbins : numpy.ndarray
        vector of centers of frequency bins from spectrogram
    timebins : numpy.ndarray
        vector of centers of time bins from spectrogram
    """
    noverlap = fft_size - step_size

    if freq_cutoffs:
        dat = butter_bandpass_filter(dat, freq_cutoffs[0], freq_cutoffs[1], samp_freq)

    # below only take [:3] from return of specgram because we don't need the image
    spect, freqbins, timebins = specgram(dat, fft_size, samp_freq, noverlap=noverlap)[
        :3
    ]

    if transform_type:
        if transform_type == "log_spect":
            spect /= spect.max()  # volume normalize to max 1
            spect = np.log10(spect)  # take log
            if thresh:
                # I know this is weird, maintaining 'legacy' behavior
                spect[spect < -thresh] = -thresh
        elif transform_type == "log_spect_plus_one":
            spect = np.log10(spect + 1)
            if thresh:
                spect[spect < thresh] = thresh
    else:
        if thresh:
            spect[
                spect < thresh
            ] = thresh  # set anything less than the threshold as the threshold

    if freq_cutoffs:
        f_inds = np.nonzero(
            (freqbins >= freq_cutoffs[0]) & (freqbins < freq_cutoffs[1])
        )[
            0
        ]  # returns tuple
        spect = spect[f_inds, :]
        freqbins = freqbins[f_inds]

    return spect, freqbins, timebins



"""validators used by attrs-based classes and by vak.parse.parse_config"""
from pathlib import Path

import toml

from .. import constants
from .. import models


def is_a_directory(instance, attribute, value):
    """check if given path is a directory"""
    if not Path(value).is_dir():
        raise NotADirectoryError(
            f"Value specified for {attribute.name} of {type(instance)} not recognized as a directory:\n"
            f"{value}"
        )


def is_a_file(instance, attribute, value):
    """check if given path is a file"""
    if not Path(value).is_file():
        raise FileNotFoundError(
            f"Value specified for {attribute.name} of {type(instance)} not recognized as a file:\n"
            f"{value}"
        )


def is_valid_model_name(instance, attribute, value):
    MODEL_NAMES = [model_name for model_name, model_builder in models.find()]
    for model_name in value:
        if model_name not in MODEL_NAMES and f"{model_name}Model" not in MODEL_NAMES:
            raise ValueError(
                f"Model {model_name} not found when importing installed models."
            )


def is_audio_format(instance, attribute, value):
    """check if valid audio format"""
    if value not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(f"{value} is not a valid format for audio files")


def is_annot_format(instance, attribute, value):
    """check if valid annotation format"""
    if value not in constants.VALID_ANNOT_FORMATS:
        raise ValueError(
            f"{value} is not a valid format for annotation files.\n"
            f"Valid formats are: {constants.VALID_ANNOT_FORMATS}"
        )


def is_spect_format(instance, attribute, value):
    """check if valid format for spectrograms"""
    if value not in constants.VALID_SPECT_FORMATS:
        raise ValueError(
            f"{value} is not a valid format for spectrogram files.\n"
            f"Valid formats are: {constants.VALID_SPECT_FORMATS}"
        )


CONFIG_DIR = Path(__file__).parent
VALID_TOML_PATH = CONFIG_DIR.joinpath("valid.toml")
with VALID_TOML_PATH.open("r") as fp:
    VALID_DICT = toml.load(fp)
VALID_SECTIONS = list(VALID_DICT.keys())
VALID_OPTIONS = {
    section: list(options.keys()) for section, options in VALID_DICT.items()
}


def are_sections_valid(config_dict, toml_path=None):
    sections = list(config_dict.keys())
    from ..cli.cli import CLI_COMMANDS  # avoid circular import

    cli_commands_besides_prep = [
        command for command in CLI_COMMANDS if command != "prep"
    ]
    sections_that_are_commands_besides_prep = [
        section for section in sections if section.lower() in cli_commands_besides_prep
    ]
    if len(sections_that_are_commands_besides_prep) == 0:
        raise ValueError(
            "did not find a section related to a vak command in config besides `prep`.\n"
            f"Sections in config were: {sections}"
        )

    if len(sections_that_are_commands_besides_prep) > 1:
        raise ValueError(
            "found multiple sections related to a vak command in config besides `prep`.\n"
            f"Those sections are: {sections_that_are_commands_besides_prep}. "
            f"Please use just one command besides `prep` per .toml configuration file"
        )

    MODEL_NAMES = [model_name for model_name, model_builder in models.find()]
    # add model names to valid sections so users can define model config in sections
    valid_sections = VALID_SECTIONS + MODEL_NAMES
    for section in sections:
        if section not in valid_sections and f"{section}Model" not in valid_sections:
            if toml_path:
                err_msg = f"section defined in {toml_path} is not valid: {section}"
            else:
                err_msg = f"section defined in toml config is not valid: {section}"
            raise ValueError(err_msg)


def are_options_valid(config_dict, section, toml_path=None):
    user_options = set(config_dict[section].keys())
    valid_options = set(VALID_OPTIONS[section])
    if not user_options.issubset(valid_options):
        invalid_options = user_options - valid_options
        if toml_path:
            err_msg = (
                f"the following options from {section} section in "
                f"the config file '{toml_path.name}' are not valid:\n{invalid_options}"
            )
        else:
            err_msg = (
                f"the following options from {section} section in "
                f"the toml config are not valid:\n{invalid_options}"
            )
        raise ValueError(err_msg)



import pytest

from hashlib import md5
from time import localtime


from vak.constants import RESULTS_DIR_PREFIX


@pytest.fixture
def previous_run_path_factory(generated_test_data_root):
    def _previous_run_path(model):
        learncurve_results_root = generated_test_data_root.joinpath(
            f"results/learncurve/audio_cbin_annot_notmat/{model}"
        )
        results_dirs = sorted(learncurve_results_root.glob(f"{RESULTS_DIR_PREFIX}*"))
        assert len(results_dirs) >= 1
        return results_dirs[-1]

    return _previous_run_path


@pytest.fixture
def random_path_factory(tmp_path):
    """factory function that returns a random path to a file that does not exist"""

    def _random_path(suffix):
        prefix = md5(str(localtime()).encode("utf-8")).hexdigest()
        return tmp_path / f"{prefix}_{suffix}"

    return _random_path



from pathlib import Path
from datetime import datetime
import warnings

import toml

from .. import config
from .. import core
from .. import logging
from ..config.parse import _load_toml_from_path
from ..config.validators import are_sections_valid


def purpose_from_toml(config_toml, toml_path=None):
    """determine "purpose" from toml config,
    i.e., the command that will be run after we ``prep`` the data.

    By convention this is the other section in the config file
    that correspond to a cli command besides '[PREP]'
    """
    # validate, make sure there aren't multiple commands in one config file first
    are_sections_valid(config_toml, toml_path=toml_path)

    from ..cli.cli import CLI_COMMANDS  # avoid circular imports

    commands_that_are_not_prep = (
        command for command in CLI_COMMANDS if command != "prep"
    )
    for command in commands_that_are_not_prep:
        section_name = (
            command.upper()
        )  # we write section names in uppercase, e.g. `[PREP]`, by convention
        if section_name in config_toml:
            return section_name.lower()  # this is the "purpose" of the file


# see https://github.com/NickleDave/vak/issues/334
SECTIONS_PREP_SHOULD_PARSE = ("PREP", "SPECT_PARAMS")


def prep(toml_path):
    """prepare datasets from vocalizations.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.
        Used to rewrite file with options determined by this function and needed for other functions

    Returns
    -------
    None

    Notes
    -----
    Saves a .csv file representing the dataset generated from data_dir.

    Datasets are used to train neural networks that segment audio files into
    vocalizations, and then predict labels for those segments.
    The function also prepares datasets so neural networks can predict the
    segmentation and annotation of vocalizations in them.
    It can also split a dataset into training, validation, and test sets,
    e.g. for benchmarking different neural network architectures.

    If the 'purpose' is set to 'train' or 'learncurve', and/or
    the duration of either the training or test set is provided,
    then the function attempts to split the dataset into training and test sets.
    A duration can also be specified for a validation set
    (used to measure performance during training).
    In these cases, the 'split' column in the .csv
    identifies which files (rows) belong to the training, test, and
    validation sets created from that Dataset.

    If the 'purpose' is set to 'predict' or 'eval',
    or no durations for any of the training sets are specified,
    then the function assumes all the vocalizations constitute a single
    dataset, and for all rows the 'split' columns for that dataset
    will be 'predict' or 'test' (respectively).
    """
    toml_path = Path(toml_path)

    # open here because need to check for `csv_path` in this function, see #314 & #333
    config_toml = _load_toml_from_path(toml_path)
    # ---- figure out purpose of config file from sections; will save csv path in that section -------------------------
    purpose = purpose_from_toml(config_toml, toml_path)
    if (
        "csv_path" in config_toml[purpose.upper()]
        and config_toml[purpose.upper()]["csv_path"] is not None
    ):
        raise ValueError(
            f"config .toml file already has a 'csv_path' option in the '{purpose.upper()}' section, "
            f"and running `prep` would overwrite that value. To `prep` a new dataset, please remove "
            f"the 'csv_path' option from the '{purpose.upper()}' section in the config file:\n{toml_path}"
        )

    # now that we've checked that, go ahead and parse the sections we want
    cfg = config.parse.from_toml_path(toml_path, sections=SECTIONS_PREP_SHOULD_PARSE)
    # notice we ignore any other option/values in the 'purpose' section,
    # see https://github.com/NickleDave/vak/issues/334 and https://github.com/NickleDave/vak/issues/314
    if cfg.prep is None:
        raise ValueError(
            f"prep called with a config.toml file that does not have a PREP section: {toml_path}"
        )

    if purpose == "predict":
        if cfg.prep.labelset is not None:
            warnings.warn(
                "config has a PREDICT section, but labelset option is specified in PREP section."
                "This would cause an error because the dataframe.from_files section will attempt to "
                f"check whether the files in the data_dir ({cfg.prep.data_dir}) have labels in "
                "labelset, even though those files don't have annotation.\n"
                "Setting labelset to None."
            )
            cfg.prep.labelset = None

    # ---- set up logging ----------------------------------------------------------------------------------------------
    timenow = datetime.now().strftime("%y%m%d_%H%M%S")
    logger = logging.get_logger(
        log_dst=cfg.prep.output_dir,
        caller="prep",
        timestamp=timenow,
        logger_name=__name__,
    )

    section = purpose.upper()
    logger.info(
        f"determined that purpose of config file is: {purpose}\n"
        f"will add 'csv_path' option to '{section}' section"
    )

    vak_df, csv_path = core.prep(
        data_dir=cfg.prep.data_dir,
        purpose=purpose,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_output_dir=cfg.prep.spect_output_dir,
        spect_params=cfg.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
        logger=logger,
    )

    # use config and section from above to add csv_path to config.toml file
    config_toml[section]["csv_path"] = str(csv_path)

    with toml_path.open("w") as fp:
        toml.dump(config_toml, fp)



from argparse import ArgumentParser
from pathlib import Path, PosixPath, WindowsPath

import numpy as np
import pandas as pd


def main(
    csv, os, audio_path_parent=None, spect_path_parent=None, annot_path_parent=None
):
    """ "move" the paths in a .csv file saved by `vak prep`,
    by keeping the filename but changing the parent path

    Parameters
    ----------
    csv : str
        prep csv filename
    os : str
        operating system to move to.
        one of {'windows', 'unix'}.
        Determines separators used for path.
    audio_path_parent : str
        new parent for paths in audio_path column
    spect_path_parent : str
        new parent for paths in spect_path column
    annot_path_parent : str
        new parent for paths in annot_path column
    """
    csv = Path(csv)
    if not csv.exists():
        raise FileNotFoundError(f"csv not found: {csv}")

    if os == "windows":
        ToPath = WindowsPath
    elif os == "unix":
        ToPath = PosixPath
    else:
        raise ValueError(f"invalid value for os: {os}")

    vak_df = pd.read_csv(csv)

    for path_column, new_parent in zip(
        ("audio_path", "spect_path", "annot_path"),
        (audio_path_parent, spect_path_parent, annot_path_parent),
    ):
        if new_parent is None:
            continue

        new_parent = ToPath(new_parent)
        if not new_parent.exists():
            raise NotADirectoryError(
                f"new parent not recognized as a directory: {new_parent}"
            )
        new_parent = new_parent.expanduser()

        paths = vak_df[path_column].values
        new_path_col = []
        for path in paths:
            old_path = Path(path)
            new_path = str(new_parent / old_path.name)
            new_path_col.append(new_path)
        vak_df[path_column] = np.array(new_path_col)

    vak_df.to_csv(csv, index=False)


TO_CHOICES = ("windows", "unix")


def get_parser():
    parser = ArgumentParser()
    parser.add_argument("csv", type=str, help='csv file that should be "moved"')
    parser.add_argument(
        "os",
        choices=TO_CHOICES,
        help=f"which OS to move paths in .csv to: {TO_CHOICES}",
    )
    parser.add_argument("--audio_path_parent", type=str)
    parser.add_argument("--spect_path_parent", type=str)
    parser.add_argument("--annot_path_parent", type=str)
    return parser


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    main(
        csv=args.csv,
        os=args.os,
        audio_path_parent=args.audio_path_parent,
        spect_path_parent=args.spect_path_parent,
        annot_path_parent=args.annot_path_parent,
    )



"""tests for vak.files.files.files module"""
import pytest

import vak.files.files


def test_find_fname():
    fname = "llb3_0003_2018_04_23_14_18_54.wav.mat"
    ext = "wav"
    out = vak.files.files.find_fname(fname, ext)
    assert out == "llb3_0003_2018_04_23_14_18_54.wav"


def test_files_from_dir_with_mat(spect_dir_mat, spect_list_mat):
    files = vak.files.files.from_dir(spect_dir_mat, "mat")
    # files.from_dir returns str not Path, need to convert fixture
    spect_list_mat = [str(spect_path) for spect_path in spect_list_mat]
    assert sorted(spect_list_mat) == sorted(files)


def test_files_from_dir_with_cbin(audio_dir_cbin, audio_list_cbin):
    files = vak.files.files.from_dir(audio_dir_cbin, "cbin")
    # files.from_dir returns str not Path, need to convert fixture
    audio_list_cbin = [str(audio_path) for audio_path in audio_list_cbin]
    assert sorted(audio_list_cbin) == sorted(files)


@pytest.mark.parametrize(
    ("dir_path", "ext"),
    [
        ("./tests/data_for_tests/source/audio_wav_annot_textgrid/AGBk/", "WAV"),
        ("./tests/data_for_tests/source/audio_wav_annot_birdsongrec/Bird0/Wave", "wav"),
    ],
)
def test_from_dir_is_case_insensitive(dir_path, ext):
    files = vak.files.files.from_dir(dir_path, ext)
    assert len(files) > 0
    assert all([str(file).endswith(ext) for file in files])


@pytest.mark.parametrize(
    ("dir_path", "ext"),
    [
        ("./tests/data_for_tests/source/audio_wav_annot_textgrid/", "WAV"),
        ("./tests/data_for_tests/source/audio_wav_annot_birdsongrec/Bird0", "wav"),
    ],
)
def test_from_dir_searches_child_dir(dir_path, ext):
    files = vak.files.files.from_dir(dir_path, ext)
    assert len(files) > 0
    assert all([str(file).endswith(ext) for file in files])



import numpy as np

from .algorithms import brute_force
from .algorithms.validate import validate_split_durations
from ..labels import from_df as labels_from_df
from ..logging import log_or_print


def train_test_dur_split_inds(
    durs,
    labels,
    labelset,
    train_dur,
    test_dur,
    val_dur=None,
    algo="brute_force",
    logger=None,
):
    """return indices to split a dataset into training, test, and validation sets of specified durations.

        Given the durations of a set of vocalizations, and labels from the annotations for those vocalizations,
        this function returns arrays of indices for splitting up the set into training, test,
        and validation sets.

        Using those indices will produce datasets that each contain instances of all labels in the set of labels.

        Parameters
        ----------
        durs : iterable
            of float. Durations of audio files.
        labels : iterable
            of numpy arrays of str or int. Labels for segments (syllables, phonemes, etc.) in audio files.
        labelset : set, list
            set of unique labels for segments in files. Used to verify that each returned array
            of indices will produce a set that contains instances of all labels found in original
            set.
        train_dur : float
            Target duration for training set, in seconds.
        test_dur : float
            Target duration for test set, in seconds.
        val_dur : float
            Target duration for validation set, in seconds. Default is None.
            If None, no indices are returned for validation set.
        algo : str
            algorithm to use. One of {'brute_force', 'inc_freq'}. Default is 'brute_force'. For more information
            on the algorithms, see the docstrings, e.g., vak.io.algorithms.brute_force
    .
        Other Parameters
        ----------------
        logger : logging.Logger
            instance created by vak.logging.get_logger. Default is None.

        Returns
        -------
        train_inds, test_inds, val_inds : numpy.ndarray
            indices to use with some array-like object to produce sets of specified durations
    """
    if len(durs) != len(labels):
        raise ValueError(
            f"length of durs, {len(durs)} does not equal length of labels, {len(labels)}"
        )

    total_dur = sum(durs)
    train_dur, val_dur, test_dur = validate_split_durations(
        train_dur, val_dur, test_dur, total_dur
    )

    total_target_dur = sum(
        [dur for dur in (train_dur, test_dur, val_dur) if dur is not None]
    )

    if total_target_dur > total_dur:
        raise ValueError(
            f"Total duration of dataset, {total_dur} seconds, is less than total target duration of "
            f"training, test, and (if specified) validation sets: {total_target_dur}"
        )

    log_or_print(
        f"Total target duration of splits: {total_target_dur} seconds. "
        f"Will be drawn from dataset with total duration: {total_dur:.3f}.",
        logger=logger,
        level="info",
    )

    if algo == "brute_force":
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset, train_dur, val_dur, test_dur
        )
    else:
        raise NotImplementedError(f"algorithm {algo} not implemented")

    return train_inds, val_inds, test_inds


def dataframe(
    vak_df, labelset, train_dur=None, test_dur=None, val_dur=None, logger=None
):
    """split a dataset of vocalizations into training, test, and (optionally) validation subsets,
    specified by their duration.

    Takes dataset represented as a pandas DataFrame and adds a 'split' column that assigns each
    row to 'train', 'val', 'test', or 'None'.

    Parameters
    ----------
    vak_df : pandas.Dataframe
        a dataset of vocalizations.
    labelset : set, list
        of str or int, set of labels for vocalizations.
    train_dur : float
        total duration of training set, in seconds. Default is None
    test_dur : float
        total duration of test set, in seconds. Default is None.
    val_dur : float
        total duration of validation set, in seconds. Default is None.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    vak_df : pandas.Dataframe
        a copy of the input dataset with a 'split' column added,
        that assigns each vocalization (row) to a subset,
        i.e., train, validation, or test.
        If the vocalization was not added to one of the subsets,
        its value for 'split' will be 'None'.

    Notes
    -----
    uses the function `vak.dataset.split.train_test_dur_split_inds` to find indices for each subset.
    """
    vak_df = (
        vak_df.copy()
    )  # don't want this function to have unexpected side effects, so return a copy
    labels = labels_from_df(vak_df)

    durs = vak_df["duration"].values
    train_inds, val_inds, test_inds = train_test_dur_split_inds(
        durs=durs,
        labels=labels,
        labelset=labelset,
        train_dur=train_dur,
        test_dur=test_dur,
        val_dur=val_dur,
        logger=logger,
    )

    # start off with all elements set to 'None'
    # so we don't have to change any that are not assigned to one of the subsets to 'None' after
    split_col = np.asarray(["None" for _ in range(len(vak_df))], dtype="object")
    split_zip = zip(["train", "val", "test"], [train_inds, val_inds, test_inds])
    for split_name, split_inds in split_zip:
        if split_inds is not None:
            split_col[split_inds] = split_name

    # add split column to dataframe
    vak_df["split"] = split_col

    return vak_df



import attr
from attr.validators import instance_of


@attr.s
class DataLoaderConfig:
    """represents options for DataLoaders specified in config.toml file

    Attributes
    ----------
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shonw to neural networks
    """

    window_size = attr.ib(converter=int, validator=instance_of(int), default=88)



"""tests for ``vak.io.audio`` module"""
from pathlib import Path

import numpy as np
import pytest

import vak.io.audio


def expected_spect_files_returned(
    spect_files_returned,
    source_audio_files_expected,
    source_audio_files_not_expected=None,
    spect_file_ext=".spect.npz",
):
    """test that ``spect_files`` returned by ``vak.io.audio.to_spect`` matches
    an expected list of spectrogram files

    Parameters
    ----------
    spect_files_returned : list
        of str, full paths to .npz array files containing spectrograms
    source_audio_files_expected : list
        of Path, source audio files expected to be used for spectrograms
    source_audio_files_not_expected : list
        of Path, source audio files expected to be used for spectrograms
    spect_file_ext : str
        extension given to array files containing spectrograms.
        Default is `.spect.npz`
    """
    assert type(spect_files_returned) == list

    spect_files_returned = [Path(spect_file) for spect_file in spect_files_returned]
    assert all([spect_file.exists() for spect_file in spect_files_returned])

    for spect_file in spect_files_returned:
        spect_dict = np.load(spect_file)
        for key in ["s", "f", "t"]:
            assert key in spect_dict
            assert type(spect_dict[key]) == np.ndarray

    # remove path so we can just compare file names
    spect_files_returned = [spect_file.name for spect_file in spect_files_returned]
    expected_spect_files = [
        source_audio_file.name + spect_file_ext
        for source_audio_file in source_audio_files_expected
    ]
    assert all(
        [
            expected_spect_file in spect_files_returned
            for expected_spect_file in expected_spect_files
        ]
    )

    if source_audio_files_not_expected is not None:
        not_expected_spect_files = [
            source_audio_file.name + spect_file_ext
            for source_audio_file in source_audio_files_not_expected
        ]
        assert all(
            [
                not_expected_spect_file not in spect_files_returned
                for not_expected_spect_file in not_expected_spect_files
            ]
        )

    return True  # if all asserts were True


def test_to_spect_audio_dir_annot_cbin_with_labelset(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    annot_list_notmat,
    labelset_notmat,
    audio_list_cbin_all_labels_in_labelset,
    audio_list_cbin_labels_not_in_labelset,
):
    """test that ``vak.io.audio.to_spect`` works
    when we point it at directory of .cbin files + give it list of annotations"""
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=audio_dir_cbin,
        audio_files=None,
        annot_list=annot_list_notmat,
        audio_annot_map=None,
        labelset=labelset_notmat,
    )
    assert expected_spect_files_returned(
        spect_files,
        audio_list_cbin_all_labels_in_labelset,
        audio_list_cbin_labels_not_in_labelset,
    )


def test_audio_dir_annot_cbin_no_labelset(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
):
    """test that ``vak.io.audio.to_spect`` works
    when we point it at directory of .cbin files + give it list of annotations
    but do not specify a set of labels"""
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=audio_dir_cbin,
        audio_files=None,
        annot_list=annot_list_notmat,
        audio_annot_map=None,
        labelset=None,
    )
    assert expected_spect_files_returned(spect_files, audio_list_cbin)


def test_audio_dir_without_annot(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
):
    """test that ``vak.io.audio.to_spect`` works
    when we point it at directory of .cbin files
    + give it list of annotations
    but do not specify a set of labels"""
    # make sure we can make a spectrograms from audio files without annotations,
    # e.g. if we're going to predict the annotations using the spectrograms
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=audio_dir_cbin,
        audio_files=None,
        annot_list=None,
        audio_annot_map=None,
        labelset=None,
    )
    assert expected_spect_files_returned(spect_files, audio_list_cbin)


def test_audio_files_cbin_annot_notmat(
    default_spect_params,
    tmp_path,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
    audio_list_cbin_all_labels_in_labelset,
    audio_list_cbin_labels_not_in_labelset,
):
    """test that ``vak.io.audio.to_spect`` works
    when we give it a list of .cbin files + give it list of annotations"""
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=None,
        audio_files=audio_list_cbin,
        annot_list=annot_list_notmat,
        audio_annot_map=None,
        labelset=labelset_notmat,
    )
    assert expected_spect_files_returned(
        spect_files,
        audio_list_cbin_all_labels_in_labelset,
        audio_list_cbin_labels_not_in_labelset,
    )


def test_audio_files_annot_cbin_no_labelset(
    default_spect_params,
    tmp_path,
    audio_list_cbin,
    annot_list_notmat,
):
    """test that ``vak.io.audio.to_spect`` works
    when we give it a list of .cbin files
    + give it list of annotations
    but do not specify a set of labels"""
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=None,
        audio_files=audio_list_cbin,
        annot_list=annot_list_notmat,
        audio_annot_map=None,
        labelset=None,
    )
    assert expected_spect_files_returned(spect_files, audio_list_cbin)


def test_audio_annot_map_cbin(
    default_spect_params,
    tmp_path,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
    audio_list_cbin_all_labels_in_labelset,
    audio_list_cbin_labels_not_in_labelset,
):
    """test that ``vak.io.audio.to_spect`` works
    when we give it a dict that maps audio files to annotations"""
    audio_annot_map = dict(zip(audio_list_cbin, annot_list_notmat))
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=None,
        audio_files=None,
        annot_list=None,
        audio_annot_map=audio_annot_map,
        labelset=labelset_notmat,
    )
    assert expected_spect_files_returned(
        spect_files,
        audio_list_cbin_all_labels_in_labelset,
        audio_list_cbin_labels_not_in_labelset,
    )


def test_audio_annot_map_cbin_no_labelset(
    default_spect_params,
    tmp_path,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """test that ``vak.io.audio.to_spect`` works
    when we give it a dict that maps audio files to annotations
    but do not give it a labelset to filter out files"""
    audio_annot_map = dict(zip(audio_list_cbin, annot_list_notmat))
    spect_files = vak.io.audio.to_spect(
        audio_format="cbin",
        spect_params=default_spect_params,
        output_dir=tmp_path,
        audio_dir=None,
        audio_files=None,
        annot_list=None,
        audio_annot_map=audio_annot_map,
        labelset=None,
    )
    assert expected_spect_files_returned(spect_files, audio_list_cbin)


def test_missing_inputs_raise(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """test that calling ``vak.io.audio.to_spect`` without one of:
    audio files, audio list, or audio files/annotations mapping
    raises a ValueError
    """
    with pytest.raises(ValueError):
        vak.io.audio.to_spect(
            audio_format="ape",
            spect_params=default_spect_params,
            output_dir=tmp_path,
            audio_dir=None,
            audio_files=None,
            annot_list=annot_list_notmat,
            audio_annot_map=None,
            labelset=labelset_notmat,
        )


def test_invalid_audio_format_raises(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """test that calling ``vak.io.audio.to_spect`` with an invalid audio format
    raises a ValueError"""
    with pytest.raises(ValueError):
        vak.io.audio.to_spect(
            audio_format="ape",
            spect_params=default_spect_params,
            output_dir=tmp_path,
            audio_dir=audio_dir_cbin,
            audio_files=None,
            annot_list=annot_list_notmat,
            audio_annot_map=None,
            labelset=labelset_notmat,
        )


def test_both_audio_dir_and_audio_files_raises(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """ "test that calling ``vak.io.audio.to_spect``
    with both audio_dir and audio_files
    raises a ValueError"""
    with pytest.raises(ValueError):
        vak.io.audio.to_spect(
            audio_format="cbin",
            spect_params=default_spect_params,
            output_dir=tmp_path,
            audio_dir=audio_dir_cbin,
            audio_files=audio_list_cbin,
            annot_list=annot_list_notmat,
            audio_annot_map=None,
            labelset=labelset_notmat,
        )


def test_both_audio_dir_and_audio_annot_map_raises(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """ "test that calling ``vak.io.audio.to_spect``
    with both audio_dir and audio_annot_map
    raises a ValueError"""
    audio_annot_map = dict(zip(audio_list_cbin, annot_list_notmat))
    with pytest.raises(ValueError):
        vak.io.audio.to_spect(
            audio_format="cbin",
            spect_params=default_spect_params,
            output_dir=tmp_path,
            audio_dir=audio_dir_cbin,
            audio_files=None,
            annot_list=None,
            audio_annot_map=audio_annot_map,
            labelset=labelset_notmat,
        )


def test_both_audio_list_and_audio_annot_map_raises(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """ "test that calling ``vak.io.audio.to_spect``
    with both audio_files and audio_annot_map
    raises a ValueError"""
    audio_annot_map = dict(zip(audio_list_cbin, annot_list_notmat))
    with pytest.raises(ValueError):
        vak.io.audio.to_spect(
            audio_format="cbin",
            spect_params=default_spect_params,
            output_dir=tmp_path,
            audio_dir=None,
            audio_files=audio_list_cbin,
            annot_list=None,
            audio_annot_map=audio_annot_map,
            labelset=labelset_notmat,
        )


def test_both_annot_list_and_audio_annot_map_raises(
    default_spect_params,
    tmp_path,
    audio_dir_cbin,
    audio_list_cbin,
    annot_list_notmat,
    labelset_notmat,
):
    """ "test that calling ``vak.io.audio.to_spect``
    with both audio_annot_map and annot_list
    raises a ValueError"""
    audio_annot_map = dict(zip(audio_list_cbin, annot_list_notmat))
    with pytest.raises(ValueError):
        vak.io.audio.to_spect(
            audio_format="cbin",
            spect_params=default_spect_params,
            output_dir=tmp_path,
            audio_dir=None,
            audio_files=None,
            annot_list=annot_list_notmat,
            audio_annot_map=audio_annot_map,
            labelset=labelset_notmat,
        )



import pytest
import torch

DEVICES = ["cpu"]
if torch.cuda.is_available():
    DEVICES.append("cuda")


@pytest.fixture(params=DEVICES)
def device(request):
    """parametrized device function,
    that returns string names of the devices
    that ``torch`` considers "available".

    causes any test using ``device`` fixture to run just once
    if only a cpu is available,
    and twice if ``torch.cuda.is_available()`` returns ``True``."""
    return request.param



"""tests for vak.cli.eval module"""
import pytest

import vak.cli.eval
import vak.config
import vak.constants
import vak.paths

from . import cli_asserts
from ..test_core.test_eval import eval_output_matches_expected


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
    ],
)
def test_eval(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    options_to_change = [
        {"section": "EVAL", "option": "output_dir", "value": str(output_dir)},
        {"section": "EVAL", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="eval",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )

    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.eval.models)

    vak.cli.eval.eval(toml_path)

    assert eval_output_matches_expected(model_config_map, output_dir)

    assert cli_asserts.log_file_created(command="eval", output_path=output_dir)



import fnmatch
from pathlib import Path
import re


def find_fname(fname, ext):
    """given a file extension, finds a filename with that extension within
    another filename. Useful to find e.g. names of audio files in
    names of spectrogram files.

    Parameters
    ----------
    fname : str
        filename to search for another filename with a specific extension
    ext : str
        extension to search for in filename

    Returns
    -------
    sub_fname : str or None

    Examples
    --------
    >>> sub_fname(fname='llb3_0003_2018_04_23_14_18_54.wav.mat', ext='wav')
    'llb3_0003_2018_04_23_14_18_54.wav'
    """
    m = re.match(f"[\S]*{ext}", fname)
    if hasattr(m, "group"):
        return m.group()
    elif m is None:
        return m


def from_dir(dir_path, ext):
    """helper function that gets all files with a given extension
    from a directory or its sub-directories.

    ``from_dir`` is case insensitive. For example, if you specify the extension
    as ``wav`` then it will return files that end in ``.wav`` or ``.WAV``.
    Similarly, if you specify ``TextGrid`` as the extension,
    it will return files that end in ``.textgrid`` and ``.TextGrid``.

    If no files with the specified extension are found in the directory,
    then the function looks in all directories within ``dir_path``
    and returns any files with the extension in those directories.
    Currently the function does not recurse, i.e., it does not look
    any deeper than one level below ``dir_path``.

    Parameters
    ----------
    dir_path : str
        path to target directory
    ext : str
        file extension to search for. E.g.,

    Returns
    -------
    files : list
        of paths to files with specified file extension

    Notes
    -----
    used by vak.io.audio.files_from_dir and vak.io.annot.files_from_dir
    """
    dir_path = Path(dir_path)
    if not dir_path.is_dir():
        raise NotADirectoryError(f"dir_path not recognized as a directory: {dir_path}")

    # use fnmatch + re to make search case-insensitive
    # adopted from:
    # https://gist.github.com/techtonik/5694830
    # https://jdhao.github.io/2019/06/24/python_glob_case_sensitivity/
    glob_pat = f"*.{ext}"
    rule = re.compile(fnmatch.translate(glob_pat), re.IGNORECASE)

    files = [
        file for file in dir_path.iterdir() if file.is_file() and rule.match(file.name)
    ]

    if len(files) == 0:
        # if we don't any files with extension, look in sub-directories
        files = []
        subdirs = [subdir for subdir in dir_path.iterdir() if subdir.is_dir()]
        for subdir in subdirs:
            files.extend(
                [
                    file
                    for file in subdir.iterdir()
                    if file.is_file() and rule.match(file.name)
                ]
            )

    if len(files) == 0:
        raise FileNotFoundError(
            f"No files with extension {ext} found in "
            f"{dir_path} or immediate sub-directories"
        )

    # TODO: use / return Path instead of strings
    return [str(file) for file in files]



"""utility functions for logging"""
import datetime
import logging
from pathlib import Path
import sys


def get_logger(log_dst, caller, logger_name, timestamp=None, level="INFO"):
    """get a logger

    Parameters
    ----------
    log_dst : str, Path
        destination directory where log file should be saved
    caller : str
        function that called get_logger, e.g. 'train' or 'prep'.
        Included in log file name.
    timestamp : str
        time stamp, included in log file name.
        If None, defaults to `datetime.now().strftime('%y%m%d_%H%M%S')`.
    logger_name : str
        name to give logger when instantiating.
    level : str
        logging level. Default is 'INFO'.

    Returns
    -------
    logger : logging.Logger
        as returned by logging.getLogger.
        Will save logs to file as well as log to stdout.
    """
    log_dst = Path(log_dst)
    if not log_dst.is_dir():
        raise NotADirectoryError(
            f"destination for log file is not a directory: {log_dst}"
        )
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    if timestamp is None:
        timestamp = datetime.now().strftime("%y%m%d_%H%M%S")
    logfile_name = log_dst.joinpath(f"{caller}_{timestamp}.log")
    logger.addHandler(logging.FileHandler(logfile_name))
    logger.addHandler(logging.StreamHandler(sys.stdout))
    return logger


def log_or_print(msg, logger=None, level=None):
    """helper function to either send a message to the logger,
    or print the message if no logger is being used

    Parameters
    ----------
    msg : str
        message to either log or print
    logger : logging.Logger
        default is None, in which case the message is printed with print function.
    level : str
        name of logging level. If logger is not None, the corresponding
        method of the logger will be used to log the message
    """
    if logger is None:
        print(msg)
    else:
        log_method = getattr(logger, level)
        log_method(msg)



"""parses [TRAIN] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .validators import is_a_directory, is_a_file, is_valid_model_name
from .. import device
from ..converters import bool_from_str, comma_separated_list, expanded_user_path


@attr.s
class TrainConfig:
    """class that represents [TRAIN] section of config.toml file

    Attributes
    ----------
    models : list
        comma-separated list of model names.
        e.g., 'models = TweetyNet, GRUNet, ConvNet'
    csv_path : str
        path to where dataset was saved as a csv.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    batch_size : int
        number of samples per batch presented to models during training.
    root_results_dir : str
        directory in which results will be created.
        The vak.cli.train function will create
        a subdirectory in this directory each time it runs.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    normalize_spectrograms : bool
        if True, use spect.utils.data.SpectScaler to normalize the spectrograms.
        Normalization is done by subtracting off the mean for each frequency bin
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model. 
        Default is None, in which case a new model is initialized. 
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results. Default is None.
    labelmap_path : str
        path to 'labelmap.json' file. Default is None.
    """

    # required
    models = attr.ib(
        converter=comma_separated_list,
        validator=[instance_of(list), is_valid_model_name],
    )
    num_epochs = attr.ib(converter=int, validator=instance_of(int))
    batch_size = attr.ib(converter=int, validator=instance_of(int))
    root_results_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)

    # optional
    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )

    results_dirname = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_directory),
        default=None,
    )

    normalize_spectrograms = attr.ib(
        converter=bool_from_str,
        validator=validators.optional(instance_of(bool)),
        default=False,
    )

    num_workers = attr.ib(validator=instance_of(int), default=2)
    device = attr.ib(validator=instance_of(str), default=device.get_default())
    shuffle = attr.ib(
        converter=bool_from_str, validator=instance_of(bool), default=True
    )

    val_step = attr.ib(
        converter=converters.optional(int),
        validator=validators.optional(instance_of(int)),
        default=None,
    )
    ckpt_step = attr.ib(
        converter=converters.optional(int),
        validator=validators.optional(instance_of(int)),
        default=None,
    )
    patience = attr.ib(
        converter=converters.optional(int),
        validator=validators.optional(instance_of(int)),
        default=None,
    )
    checkpoint_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )
    spect_scaler_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )
    labelmap_path = attr.ib(
        converter=converters.optional(expanded_user_path),
        validator=validators.optional(is_a_file),
        default=None,
    )


import numpy as np
import pandas as pd
import random
import torch
from torchvision.datasets.vision import VisionDataset

from .. import annotation
from .. import files
from .. import io
from .. import labeled_timebins
from .. import validators


class WindowDataset(VisionDataset):
    """Dataset class that represents all possible windows
     of a fixed width from a set of spectrograms.
     The underlying dataset consists of spectrograms
     of vocalizations and annotations for those vocalizations.

    Returns windows from the spectrograms, along with labels for each
    time bin in the window, derived from the annotations.

    Abstraction that enables training on a dataset of a specified duration.

    Attributes
    ----------
    root : str, Path
        path to a .csv file that represents the dataset.
        Name 'root' is used for consistency with torchvision.datasets
    x_inds : numpy.ndarray
        vector [1,2,3,...,n] that indexes the number of windows
        in entire dataset. See ``Notes`` for more detail.
    spect_id_vector : numpy.ndarray
        represents the 'id' of any spectrogram,
        i.e., the index into spect_paths that will let us load it
    spect_inds_vector : numpy.ndarray
        valid indices of windows we can grab from each spectrogram
    spect_paths : numpy.ndarray
        column from DataFrame that represents dataset,
        consisting of paths to files containing spectrograms as arrays
    annots : list
        of crowsetta.Annotation instances,
        loaded from from DataFrame that represents dataset, using vak.annotation.from_df.
    labelmap : dict
        that maps labels from dataset to a series of consecutive integer.
        To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
    timebin_dur : float
        duration of a single time bin in spectrograms.
    window_size : int
        number of time bins in windows that will be taken from spectrograms
    spect_key : str
        key to access spectograms in array files. Default is 's'.
    timebins_key : str
        key to access time bin vector in array files. Default is 't'.
    transform : callable
        A function/transform that takes in a numpy array or torch Tensor
        and returns a transformed version. E.g, vak.transforms.StandardizeSpect
        Default is None.
    target_transform : callable
        A function/transform that takes in the target and transforms it.

    Notes
    -----
    This class uses three vectors to represent
    a dataset of windows from spectrograms, without actually loading
    all the spectrograms and concatenating them into one big array.
    The three vectors correspond to this imaginary, unloaded big matrix:
    (1) ``spect_id_vector`` that represents the "id" of any spectrogram in this matrix,
    i.e., the index into ``spect_path`` of the dataset .csv that will let us load it, and
    (2) ``spect_inds_vector`` where the elements represents valid indices of windows
    we can grab from each spectrogram. Valid indices are any up to the index n, where
    n = number of time bins in this spectrogram - number of time bins in our window
    (because if we tried to go past that the window would go past the edge of the
    spectrogram).
    (3) ``x_inds`` is our 'training set' vector, just a set
    of indices (0, 1, ..., m) where m is the length of vectors (1) and (2).
    When we want to grab a batch of size b of windows, we get b indices from x,
    and then index into vectors (1) and (2) so we know which spectrogram files to
    load, and which windows to grab from each spectrogram.
    In terms of implementation: when a ``torch.DataLoader`` calls ``__getitem__``
    with ``idx``, we index into this vector that maps to valid windows, as
    determined by ``spect_id_vector`` and ``spect_inds_vector``

    This could be done without ``x_inds``, but ``x_inds`` lets us
    more explicitly return a ``__len__`` for the dataset.
    """

    # class attribute, constant used by several methods
    # with x_inds, to mark invalid starting indices for windows
    INVALID_WINDOW_VAL = -1

    VALID_SPLITS = ("train", "val" "test", "all")

    def __init__(
        self,
        root,
        x_inds,
        spect_id_vector,
        spect_inds_vector,
        spect_paths,
        annots,
        labelmap,
        timebin_dur,
        window_size,
        spect_key="s",
        timebins_key="t",
        transform=None,
        target_transform=None,
    ):
        """initialize a WindowDataset instance

        Parameters
        ----------
        root : str, Path
            path to a .csv file that represents the dataset.
            Name 'root' is used for consistency with torchvision.datasets
        x_inds : numpy.ndarray
            indices of each window in the dataset. The value at x[0]
            represents the start index of the first window; using that
            value, we can index into spect_id_vector to get the path
            of the spectrogram file to load, and we can index into
            spect_inds_vector to index into the spectrogram itself
            and get the window.
        spect_id_vector : numpy.ndarray
            represents the 'id' of any spectrogram,
            i.e., the index into spect_paths that will let us load it
        spect_inds_vector : numpy.ndarray
            same length as spect_id_vector but values represent
            indices within each spectrogram.
        spect_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to files containing spectrograms as arrays
        annots : list
            of crowsetta.Annotation instances,
            loaded from from DataFrame that represents dataset, using vak.annotation.from_df.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        timebin_dur : float
            duration of a single time bin in spectrograms.
        window_size : int
            number of time bins in windows that will be taken from spectrograms
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        transform : callable
            A function/transform that takes in a numpy array or torch Tensor
            and returns a transformed version. E.g, vak.transforms.StandardizeSpect
            Default is None.
        target_transform : callable
            A function/transform that takes in the target and transforms it.
        """
        super(WindowDataset, self).__init__(
            root, transform=transform, target_transform=target_transform
        )
        self.x_inds = x_inds
        self.spect_id_vector = spect_id_vector
        self.spect_inds_vector = spect_inds_vector
        self.spect_paths = spect_paths
        self.spect_key = spect_key
        self.timebins_key = timebins_key
        self.annots = annots
        self.labelmap = labelmap
        self.timebin_dur = timebin_dur
        if "unlabeled" in self.labelmap:
            self.unlabeled_label = self.labelmap["unlabeled"]
        else:
            # if there is no "unlabeled label" (e.g., because all segments have labels)
            # just assign dummy value that will end up getting replaced by actual labels by label_timebins()
            self.unlabeled_label = 0
        self.window_size = window_size

        tmp_x_ind = 0
        one_x, _ = self.__getitem__(tmp_x_ind)
        # used by vak functions that need to determine size of window,
        # e.g. when initializing a neural network model
        self.shape = one_x.shape

    def __get_window_labelvec(self, idx):
        """helper function that gets batches of training pairs,
        given indices into dataset

        Parameters
        ----------
        idx : integer
            index into dataset

        Returns
        -------
        window : numpy.ndarray
            window from spectrograms
        labelvec : numpy.ndarray
            vector of labels for each timebin in window from spectrogram
        """
        x_ind = self.x_inds[idx]
        spect_id = self.spect_id_vector[x_ind]
        window_start_ind = self.spect_inds_vector[x_ind]

        spect_path = self.spect_paths[spect_id]
        spect_dict = files.spect.load(spect_path)
        spect = spect_dict[self.spect_key]
        timebins = spect_dict[self.timebins_key]

        annot = self.annots[
            spect_id
        ]  # "annot id" == spect_id if both were taken from rows of DataFrame
        lbls_int = [self.labelmap[lbl] for lbl in annot.seq.labels]
        lbl_tb = labeled_timebins.label_timebins(
            lbls_int,
            annot.seq.onsets_s,
            annot.seq.offsets_s,
            timebins,
            unlabeled_label=self.unlabeled_label,
        )

        window = spect[:, window_start_ind : window_start_ind + self.window_size]
        labelvec = lbl_tb[window_start_ind : window_start_ind + self.window_size]

        return window, labelvec

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        window, labelvec = self.__get_window_labelvec(idx)

        if self.transform is not None:
            window = self.transform(window)

        if self.target_transform is not None:
            labelvec = self.target_transform(labelvec)

        return window, labelvec

    def __len__(self):
        """number of batches"""
        return len(self.x_inds)

    def duration(self):
        """duration of WindowDataset, in seconds"""
        return self.spect_inds_vector.shape[-1] * self.timebin_dur

    @staticmethod
    def crop_spect_vectors_keep_classes(
        lbl_tb,
        spect_id_vector,
        spect_inds_vector,
        x_inds,
        crop_dur,
        timebin_dur,
        labelmap,
        window_size,
    ):
        """crop spect_id_vector and spect_ind_vector to a target duration
        while making sure that all classes are present in the cropped
        vectors

        Parameters
        ----------
        lbl_tb : numpy.ndarray
            labeled timebins, where labels are from the set of values in labelmap.
        spect_id_vector : numpy.ndarray
            represents the 'id' of any spectrogram,
            i.e., the index into spect_paths that will let us load it
        spect_inds_vector : numpy.ndarray
            same length as spect_id_vector but values represent
            indices within each spectrogram.
        x_inds : numpy.ndarray
            indices of each window in the dataset. The value at x[0]
            represents the start index of the first window; using that
            value, we can index into spect_id_vector to get the path
            of the spectrogram file to load, and we can index into
            spect_inds_vector to index into the spectrogram itself
            and get the window.
        crop_dur : float
            duration to which dataset should be "cropped". Default is None,
            in which case entire duration of specified split will be used.
        timebin_dur : float
            duration of a single time bin in spectrograms. Default is None.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integers.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.

        Returns
        -------
        spect_id_cropped : numpy.ndarray
            spect_id_vector after cropping
        spect_inds_cropped : numpy.ndarray
            spect_inds_vector after cropping
        x_inds_updated : numpy.ndarray
            x_inds_vector with starting indices of windows that are invalid
            after the cropping now set to WindowDataset.INVALID_WINDOW_VAL
            so they will be removed
        """
        lbl_tb = validators.column_or_1d(lbl_tb)
        spect_id_vector = validators.column_or_1d(spect_id_vector)
        spect_inds_vector = validators.column_or_1d(spect_inds_vector)
        x_inds = validators.column_or_1d(x_inds)

        lens = (
            lbl_tb.shape[-1],
            spect_id_vector.shape[-1],
            spect_inds_vector.shape[-1],
            x_inds.shape[-1],
        )
        uniq_lens = set(lens)
        if len(uniq_lens) != 1:
            raise ValueError(
                "lbl_tb, spect_id_vector, spect_inds_vector, and x_inds should all "
                "have the same length, but did not find one unique length. "
                "Lengths of lbl_tb, spect_id_vector, spect_inds_vector, and x_inds_vector "
                f"were: {lens}"
            )

        cropped_length = np.round(crop_dur / timebin_dur).astype(int)

        if spect_id_vector.shape[-1] == cropped_length:
            return spect_id_vector, spect_inds_vector, x_inds

        elif spect_id_vector.shape[-1] < cropped_length:
            raise ValueError(
                f"arrays have length {spect_id_vector.shape[-1]} "
                f"that is shorter than correct length, {cropped_length}, "
                f"(= target duration {crop_dur} / duration of timebins, {timebin_dur})."
            )

        elif spect_id_vector.shape[-1] > cropped_length:
            classes = np.asarray(sorted(list(labelmap.values())))

            # try cropping off the end first
            lbl_tb_cropped = lbl_tb[:cropped_length]

            if np.array_equal(np.unique(lbl_tb_cropped), classes):
                x_inds[cropped_length:] = WindowDataset.INVALID_WINDOW_VAL
                return (
                    spect_id_vector[:cropped_length],
                    spect_inds_vector[:cropped_length],
                    x_inds,
                )

            # try truncating off the front instead
            lbl_tb_cropped = lbl_tb[-cropped_length:]
            if np.array_equal(np.unique(lbl_tb_cropped), classes):
                # set every index *up to but not including* the first valid window start to "invalid"
                x_inds[:-cropped_length] = WindowDataset.INVALID_WINDOW_VAL
                # also need to 'reset' the indexing so it starts at 0. First find current minimum index value
                min_x_ind = x_inds[x_inds != WindowDataset.INVALID_WINDOW_VAL].min()
                # Then set min x ind to 0, min x ind + 1 to 1, min ind + 2 to 2, ...
                x_inds[x_inds != WindowDataset.INVALID_WINDOW_VAL] = (
                    x_inds[x_inds != WindowDataset.INVALID_WINDOW_VAL] - min_x_ind
                )
                return (
                    spect_id_vector[-cropped_length:],
                    spect_inds_vector[-cropped_length:],
                    x_inds,
                )

            # try cropping silences
            # This is done by seeking segments > window_size + 2 bins and removing from them
            # When using this option we do not crop the spect vector sizes

            # Ignored data is defined as data that does not appear in any training window.
            # This means that there are 3 distinct cases:
            # 1. Silence in the beginning of a file
            # 2. Silence in the middle of a file
            # 3. Silence at the end of the file
            # (here 'end' means the segment prior to the last window_size bins in the file because
            # Those are not used as startpoints of a training window)

            # assigining WindowDataset.INVALID_WINDOW_VAL to x_inds segments
            # in these 3 cases will cause data to be ignored with
            # durations that depend on wether the segments touch the ends of files
            # because we do not ignore non-silence segments.

            # first identify all silence segments larger than the window duration + 2
            if "unlabeled" in labelmap:
                unlabeled = labelmap["unlabeled"]
            else:
                raise ValueError(
                    "was not able to crop spect vectors to specified duration; "
                    "could not crop from start or end, and there are no unlabeled segments "
                    "that could be used to further crop"
                )
            valid_unlabeled = np.logical_and(
                lbl_tb == unlabeled, x_inds != WindowDataset.INVALID_WINDOW_VAL
            )
            unlabeled_diff = np.diff(np.concatenate([[0], valid_unlabeled, [0]]))
            unlabeled_onsets = np.where(unlabeled_diff == 1)[0]
            unlabeled_offsets = np.where(unlabeled_diff == -1)[0]
            unlabeled_durations = unlabeled_offsets - unlabeled_onsets
            N_PAD_BINS = 2
            unlabeled_onsets = unlabeled_onsets[
                unlabeled_durations >= window_size + N_PAD_BINS
            ]
            unlabeled_offsets = unlabeled_offsets[
                unlabeled_durations >= window_size + N_PAD_BINS
            ]
            unlabeled_durations = unlabeled_durations[
                unlabeled_durations >= window_size + N_PAD_BINS
            ]
            # indicate silences in the beginning of files
            border_onsets = (
                np.concatenate([[WindowDataset.INVALID_WINDOW_VAL], x_inds])[
                    unlabeled_onsets
                ]
                == WindowDataset.INVALID_WINDOW_VAL
            )
            # indicate silences at the end of files
            border_offsets = (
                np.concatenate([x_inds, [WindowDataset.INVALID_WINDOW_VAL]])[
                    unlabeled_offsets + 1
                ]
                == WindowDataset.INVALID_WINDOW_VAL
            )

            # This is how much data can be ignored from each silence segment without ignoring the end of file windows
            num_potential_ignored_data_bins = (
                unlabeled_durations
                - (window_size + N_PAD_BINS)
                + window_size * border_onsets
            )

            num_bins_to_crop = len(lbl_tb) - cropped_length
            if sum(num_potential_ignored_data_bins) < num_bins_to_crop:
                # This is how much data can be ignored from each silence segment including the end of file windows
                num_potential_ignored_data_bins = (
                    unlabeled_durations
                    - (window_size - N_PAD_BINS)
                    + window_size * (border_onsets + border_offsets)
                )
            else:
                border_offsets[:] = False

            # Second we find a ~random combination to remove
            crop_more = 0
            if sum(num_potential_ignored_data_bins) < num_bins_to_crop:
                # if we will still need to crop more we will do so from non-silence segments
                crop_more = num_bins_to_crop - sum(num_potential_ignored_data_bins) + 1
                num_bins_to_crop = sum(num_potential_ignored_data_bins) - 1

            segment_ind = np.arange(len(num_potential_ignored_data_bins))
            random.shuffle(segment_ind)
            last_ind = np.where(
                np.cumsum(num_potential_ignored_data_bins[segment_ind])
                >= num_bins_to_crop
            )[0][0]
            bins_to_ignore = np.array([], dtype=int)
            for cnt in range(last_ind):
                if border_onsets[segment_ind[cnt]]:  # remove silences at file onsets
                    bins_to_ignore = np.concatenate(
                        [
                            bins_to_ignore,
                            np.arange(
                                unlabeled_onsets[segment_ind[cnt]],
                                unlabeled_offsets[segment_ind[cnt]] - 1,
                            ),
                        ]
                    )
                elif border_offsets[
                    segment_ind[cnt]
                ]:  # remove silences at file offsets
                    bins_to_ignore = np.concatenate(
                        [
                            bins_to_ignore,
                            np.arange(
                                unlabeled_onsets[segment_ind[cnt]] + 1,
                                unlabeled_offsets[segment_ind[cnt]],
                            ),
                        ]
                    )
                else:  # remove silences within the files
                    bins_to_ignore = np.concatenate(
                        [
                            bins_to_ignore,
                            np.arange(
                                unlabeled_onsets[segment_ind[cnt]] + 1,
                                unlabeled_offsets[segment_ind[cnt]] - 1,
                            ),
                        ]
                    )
            left_to_crop = (
                num_bins_to_crop
                - sum(num_potential_ignored_data_bins[segment_ind[:last_ind]])
                - border_onsets[segment_ind[last_ind]] * window_size
            )
            if border_onsets[segment_ind[last_ind]]:
                bins_to_ignore = np.concatenate(
                    [
                        bins_to_ignore,
                        np.arange(
                            unlabeled_onsets[segment_ind[last_ind]],
                            unlabeled_onsets[segment_ind[last_ind]] + left_to_crop,
                        ),
                    ]
                )
            elif border_offsets[segment_ind[last_ind]]:
                if (
                    left_to_crop
                    < num_potential_ignored_data_bins[segment_ind[last_ind]]
                    - window_size
                ):
                    bins_to_ignore = np.concatenate(
                        [
                            bins_to_ignore,
                            np.arange(
                                unlabeled_onsets[segment_ind[last_ind]] + 1,
                                unlabeled_onsets[segment_ind[last_ind]] + left_to_crop,
                            ),
                        ]
                    )
                else:
                    bins_to_ignore = np.concatenate(
                        [
                            bins_to_ignore,
                            np.arange(
                                unlabeled_onsets[segment_ind[last_ind]] + 1,
                                unlabeled_onsets[segment_ind[last_ind]]
                                + left_to_crop
                                - window_size,
                            ),
                        ]
                    )
            else:
                bins_to_ignore = np.concatenate(
                    [
                        bins_to_ignore,
                        np.arange(
                            unlabeled_onsets[segment_ind[last_ind]] + 1,
                            unlabeled_onsets[segment_ind[last_ind]] + left_to_crop,
                        ),
                    ]
                )

            x_inds[bins_to_ignore] = WindowDataset.INVALID_WINDOW_VAL

            # we may still need to crop. Try doing it from the beginning of the dataset
            if (
                crop_more > 0
            ):  # This addition can lead to imprecision but only in cases where we ask for very small datasets
                if crop_more > sum(x_inds != WindowDataset.INVALID_WINDOW_VAL):
                    raise ValueError(
                        "was not able to crop spect vectors to specified duration "
                        "in a way that maintained all classes in dataset"
                    )
                extra_bins = x_inds[x_inds != WindowDataset.INVALID_WINDOW_VAL][
                    :crop_more
                ]
                bins_to_ignore = np.concatenate([bins_to_ignore, extra_bins])
                x_inds[bins_to_ignore] = WindowDataset.INVALID_WINDOW_VAL

            if np.array_equal(
                np.unique(lbl_tb[np.setdiff1d(np.arange(len(lbl_tb)), bins_to_ignore)]),
                classes,
            ):
                return spect_id_vector, spect_inds_vector, x_inds

        raise ValueError(
            "was not able to crop spect vectors to specified duration "
            "in a way that maintained all classes in dataset"
        )

    @staticmethod
    def n_time_bins_spect(spect_path, spect_key="s"):
        """get number of time bins in a spectrogram,
        given a path to the array file containing that spectrogram

        Parameters
        ----------
        spect_path : str, pathlib.Path
            path to an array file containing a spectrogram and associated arrays.
        spect_key : str
            key to access spectograms in array files. Default is 's'.

        Returns
        -------
        spect.shape[-1], the number of time bins in the spectrogram.
        Assumes spectrogram is a 2-d matrix where rows are frequency bins,
        and columns are time bins.
        """
        spect = files.spect.load(spect_path)[spect_key]
        return spect.shape[-1]

    @staticmethod
    def spect_vectors_from_df(
        df,
        split,
        window_size,
        spect_key="s",
        timebins_key="t",
        crop_dur=None,
        timebin_dur=None,
        labelmap=None,
    ):
        """get spect_id_vector and spect_ind_vector from a dataframe
        that represents a dataset of vocalizations.
        See WindowDataset class docstring for
        detailed explanation of these vectors.

        Parameters
        ----------
        df : pandas.DataFrame
            that represents a dataset of vocalizations.
        window_size : int
            number of time bins in windows that will be taken from spectrograms
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        crop_dur : float
            duration to which dataset should be "cropped". Default is None,
            in which case entire duration of specified split will be used.
        timebin_dur : float
            duration of a single time bin in spectrograms. Default is None.
            Used when "cropping" dataset with crop_dur and required if a
            value is specified for that parameter.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integers.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
            Used when "cropping" dataset with crop_dur and required if a
            value is specified for that parameter.

        Returns
        -------
        spect_id_vector : numpy.ndarray
            represents the 'id' of any spectrogram,
            i.e., the index into spect_paths that will let us load it
        spect_inds_vector : numpy.ndarray
            valid indices of windows we can grab from each spectrogram
        x_inds_updated : numpy.ndarray
            x_inds_vector with starting indices of windows that are invalid
            after the cropping now set to WindowDataset.INVALID_WINDOW_VAL
            so they will be removed
        """
        if crop_dur is not None and timebin_dur is None:
            raise ValueError("must provide timebin_dur when specifying crop_dur")

        if crop_dur is not None and labelmap is None:
            raise ValueError("must provide labelmap when specifying crop_dur")

        if split not in WindowDataset.VALID_SPLITS:
            raise ValueError(
                f"invalid value for split: {split}. Valid split names are: {WindowDataset.VALID_SPLITS}"
            )

        if crop_dur is not None and timebin_dur is not None:
            crop_to_dur = True
            crop_dur = float(crop_dur)
            timebin_dur = float(timebin_dur)
            annots = annotation.from_df(df)
            if "unlabeled" in labelmap:
                unlabeled_label = labelmap["unlabeled"]
            else:
                # if there is no "unlabeled label" (e.g., because all segments have labels)
                # just assign dummy value that will end up getting replaced by actual labels by label_timebins()
                unlabeled_label = 0
        else:
            crop_to_dur = False

        if "split" == "all":
            pass  # use all rows, don't select by split
        else:
            if split not in df.split.unique().tolist():
                raise ValueError(
                    f"split {split} not found in dataframe. Split are: {df.split.unique().tolist()}"
                )
            df = df[df["split"] == split]

        spect_paths = df["spect_path"].values

        spect_id_vector = []
        spect_inds_vector = []
        x_inds = []
        total_tb = 0

        if crop_to_dur:
            lbl_tb = []
            spect_annot_map = annotation.source_annot_map(spect_paths, annots)
            for ind, (spect_path, annot) in enumerate(spect_annot_map.items()):
                spect_dict = files.spect.load(spect_path)
                n_tb_spect = spect_dict[spect_key].shape[-1]

                spect_id_vector.append(np.ones((n_tb_spect,), dtype=np.int64) * ind)
                spect_inds_vector.append(np.arange(n_tb_spect))

                valid_x_inds = np.arange(total_tb, total_tb + n_tb_spect)
                last_valid_window_ind = total_tb + n_tb_spect - window_size
                valid_x_inds[
                    valid_x_inds > last_valid_window_ind
                ] = WindowDataset.INVALID_WINDOW_VAL
                x_inds.append(valid_x_inds)

                total_tb += n_tb_spect

                lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
                timebins = spect_dict[timebins_key]
                lbl_tb.append(
                    labeled_timebins.label_timebins(
                        lbls_int,
                        annot.seq.onsets_s,
                        annot.seq.offsets_s,
                        timebins,
                        unlabeled_label=unlabeled_label,
                    )
                )

            spect_id_vector = np.concatenate(spect_id_vector)
            spect_inds_vector = np.concatenate(spect_inds_vector)
            lbl_tb = np.concatenate(lbl_tb)
            x_inds = np.concatenate(x_inds)

            (
                spect_id_vector,
                spect_inds_vector,
                x_inds,
            ) = WindowDataset.crop_spect_vectors_keep_classes(
                lbl_tb,
                spect_id_vector,
                spect_inds_vector,
                x_inds,
                crop_dur,
                timebin_dur,
                labelmap,
                window_size,
            )

        else:  # crop_to_dur is False
            for ind, spect_path in enumerate(spect_paths):
                n_tb_spect = WindowDataset.n_time_bins_spect(spect_path, spect_key)

                spect_id_vector.append(np.ones((n_tb_spect,), dtype=np.int64) * ind)
                spect_inds_vector.append(np.arange(n_tb_spect))

                valid_x_inds = np.arange(total_tb, total_tb + n_tb_spect)
                last_valid_window_ind = total_tb + n_tb_spect - window_size
                valid_x_inds[
                    valid_x_inds > last_valid_window_ind
                ] = WindowDataset.INVALID_WINDOW_VAL
                x_inds.append(valid_x_inds)

                total_tb += n_tb_spect

            spect_id_vector = np.concatenate(spect_id_vector)
            spect_inds_vector = np.concatenate(spect_inds_vector)
            x_inds = np.concatenate(x_inds)

        x_inds = x_inds[x_inds != WindowDataset.INVALID_WINDOW_VAL]
        return spect_id_vector, spect_inds_vector, x_inds

    @staticmethod
    def spect_vectors_from_csv(
        csv_path,
        split,
        window_size,
        spect_key="s",
        timebins_key="t",
        crop_dur=None,
        timebin_dur=None,
        labelmap=None,
    ):
        """get spect_id_vector and spect_ind_vector from a
        .csv file that represents a dataset of vocalizations.
        See WindowDataset class docstring for
        detailed explanation of these vectors.

        Parameters
        ----------
        csv_path : str, Path
            path to csv that represents dataset.
        split : str
            name of split from dataset to use
        window_size : int
            number of time bins in windows that will be taken from spectrograms
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integers.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        crop_dur : float
            duration to which dataset should be "cropped". Default is None,
            in which case entire duration of specified split will be used.
        timebin_dur : float
            duration of a single time bin in spectrograms. Default is None.
            Used when "cropping" dataset with crop_dur and required if a
            value is specified for that parameter.

        Returns
        -------
        spect_id_vector : numpy.ndarray
            represents the 'id' of any spectrogram,
            i.e., the index into spect_paths that will let us load it
        spect_inds_vector : numpy.ndarray
            valid indices of windows we can grab from each spectrogram
        x_inds_updated : numpy.ndarray
            x_inds_vector with starting indices of windows that are invalid
            after the cropping now set to WindowDataset.INVALID_WINDOW_VAL
            so they will be removed
        """
        if split not in WindowDataset.VALID_SPLITS:
            raise ValueError(
                f"invalid value for split: {split}. Valid split names are: {WindowDataset.VALID_SPLITS}"
            )

        df = pd.read_csv(csv_path)

        if split not in df.split.unique().tolist():
            raise ValueError(
                f"split {split} not found in dataset in csv: {csv_path}. "
                f"Splits are: {df.split.unique().tolist()}"
            )

        return WindowDataset.spect_vectors_from_df(
            df,
            split,
            window_size,
            spect_key,
            timebins_key,
            crop_dur,
            timebin_dur,
            labelmap,
        )

    @classmethod
    def from_csv(
        cls,
        csv_path,
        split,
        labelmap,
        window_size,
        spect_key="s",
        timebins_key="t",
        spect_id_vector=None,
        spect_inds_vector=None,
        x_inds=None,
        transform=None,
        target_transform=None,
    ):
        """given a path to a csv representing a dataset,
        returns an initialized WindowDataset.

        Parameters
        ----------
        csv_path : str, Path
            path to csv that represents dataset.
        split : str
            name of split from dataset to use
        labelmap : dict
            that maps labels from dataset to a series of consecutive integers.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        window_size : int
            number of time bins in windows that will be taken from spectrograms
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        spect_id_vector : numpy.ndarray
            represents the 'id' of any spectrogram,
            i.e., the index into spect_paths that will let us load it
        spect_inds_vector : numpy.ndarray
            valid indices of windows we can grab from each spectrogram
        x_inds : numpy.ndarray
            indices of each window in the dataset. The value at x[0]
            represents the start index of the first window; using that
            value, we can index into spect_id_vector to get the path
            of the spectrogram file to load, and we can index into
            spect_inds_vector to index into the spectrogram itself
            and get the window.
        transform : callable
            A function/transform that takes in a numpy array
            and returns a transformed version. E.g, a SpectScaler instance.
            Default is None.
        target_transform : callable
            A function/transform that takes in the target and transforms it.

        Returns
        -------
        initialized instance of WindowDataset
        """
        if any(
            [vec is not None for vec in [spect_id_vector, spect_inds_vector, x_inds]]
        ):
            if not all(
                [
                    vec is not None
                    for vec in [spect_id_vector, spect_inds_vector, x_inds]
                ]
            ):

                raise ValueError(
                    "if any of the following parameters are specified, they all must be specified: "
                    "spect_id_vector, spect_inds_vector, x_inds"
                )

        if all(
            [vec is not None for vec in [spect_id_vector, spect_inds_vector, x_inds]]
        ):
            for vec_name, vec in zip(
                ["spect_id_vector", "spect_inds_vector", "x_inds"],
                [spect_id_vector, spect_inds_vector, x_inds],
            ):
                if not type(vec) is np.ndarray:
                    raise TypeError(
                        f"{vec_name} must be a numpy.ndarray but type was: {type(spect_id_vector)}"
                    )

            spect_id_vector = validators.column_or_1d(spect_id_vector)
            spect_inds_vector = validators.column_or_1d(spect_inds_vector)
            x_inds = validators.column_or_1d(x_inds)

            if spect_id_vector.shape[-1] != spect_inds_vector.shape[-1]:
                raise ValueError(
                    "spect_id_vector and spect_inds_vector should be same length, but "
                    f"spect_id_vector.shape[-1] is {spect_id_vector.shape[-1]} and "
                    f"spect_inds_vector.shape[-1] is {spect_inds_vector.shape[-1]}."
                )

        df = pd.read_csv(csv_path)
        if not df["split"].str.contains(split).any():
            raise ValueError(f"split {split} not found in dataset in csv: {csv_path}")
        else:
            df = df[df["split"] == split]
        spect_paths = df["spect_path"].values

        if all([vec is None for vec in [spect_id_vector, spect_inds_vector, x_inds]]):
            # see Notes in class docstring to understand what these vectors do
            spect_id_vector, spect_inds_vector, x_inds = cls.spect_vectors_from_df(
                df, split, window_size
            )

        annots = annotation.from_df(df)
        timebin_dur = io.dataframe.validate_and_get_timebin_dur(df)

        # note that we set "root" to csv path
        return cls(
            csv_path,
            x_inds,
            spect_id_vector,
            spect_inds_vector,
            spect_paths,
            annots,
            labelmap,
            timebin_dur,
            window_size,
            spect_key,
            timebins_key,
            transform,
            target_transform,
        )



"""tests for vak.cli.train module"""
import pytest

import vak.config
import vak.constants
import vak.paths
import vak.cli.train

from . import cli_asserts
from ..test_core.test_train import train_output_matches_expected


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format",
    [
        ("cbin", None, "notmat"),
        ("wav", None, "birdsong-recognition-dataset"),
        (None, "mat", "yarden"),
    ],
)
def test_train(
    audio_format, spect_format, annot_format, specific_config, tmp_path, model, device
):
    root_results_dir = tmp_path.joinpath("test_train_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "TRAIN",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "TRAIN", "option": "device", "value": device},
    ]

    toml_path = specific_config(
        config_type="train",
        model=model,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        options_to_change=options_to_change,
    )

    vak.cli.train.train(toml_path)

    cfg = vak.config.parse.from_toml_path(toml_path)
    model_config_map = vak.config.models.map_from_path(toml_path, cfg.train.models)
    results_path = sorted(root_results_dir.glob(f"{vak.constants.RESULTS_DIR_PREFIX}*"))
    assert len(results_path) == 1
    results_path = results_path[0]

    assert train_output_matches_expected(cfg, model_config_map, results_path)

    assert cli_asserts.toml_config_file_copied_to_results_path(results_path, toml_path)
    assert cli_asserts.log_file_created(command="train", output_path=results_path)



import tempfile

import pandas as pd
import pytest

import vak.converters
import vak.core.learncurve.train_dur_csv_paths
import vak.csv
import vak.io.dataframe
import vak.labels
import vak.paths

SPECT_KEY = "s"
TIMEBINS_KEY = "t"


@pytest.fixture
def tmp_previous_run_path_factory(tmp_path):
    def _tmp_previous_run_path(
        train_set_durs, num_replicates, dataset_csv_path_filename="fake_data_prep.csv"
    ):
        results_path = vak.paths.generate_results_dir_name_as_path(tmp_path)
        results_path.mkdir()
        csv_path = tmp_path.joinpath("data_dir").joinpath(dataset_csv_path_filename)

        for train_set_dur in train_set_durs:
            path_this_train_dur = results_path.joinpath(
                vak.core.learncurve.train_dur_csv_paths.train_dur_dirname(train_set_dur)
            )
            path_this_train_dur.mkdir()
            for replicate_num in range(1, num_replicates + 1):
                path_this_replicate = path_this_train_dur.joinpath(
                    vak.core.learncurve.train_dur_csv_paths.replicate_dirname(
                        replicate_num
                    )
                )
                path_this_replicate.mkdir()
                fake_subset_csv_path = path_this_replicate.joinpath(
                    vak.core.learncurve.train_dur_csv_paths.subset_csv_filename(
                        csv_path, train_set_dur, replicate_num
                    )
                )
                # make fake subset csv path exist
                with fake_subset_csv_path.open("w") as fp:
                    fp.write("")
                assert fake_subset_csv_path.exists()

        return results_path

    return _tmp_previous_run_path


@pytest.mark.parametrize(
    "train_set_durs, num_replicates",
    [
        ([4, 6], 2),
        ([30, 45, 60, 120, 180], 10),
    ],
)
def test_dict_from_dir(tmp_previous_run_path_factory, train_set_durs, num_replicates):
    """note: this test specifically tests that
    csv paths are sorted numerically by replicate number,
    not alphabetically.
    See https://github.com/NickleDave/vak/issues/340

    this is the main reason for factoring out the helper function ``_dict_from_dir``
    """
    previous_run_path = tmp_previous_run_path_factory(train_set_durs, num_replicates)

    train_dur_csv_paths = vak.core.learncurve.train_dur_csv_paths._dict_from_dir(
        previous_run_path
    )

    assert isinstance(train_dur_csv_paths, dict)
    assert sorted(train_dur_csv_paths.keys()) == train_set_durs
    for train_set_dur, csv_list in train_dur_csv_paths.items():
        replicate_nums = [int(csv.parent.name.split("_")[-1]) for csv in csv_list]
        assert replicate_nums == list(range(1, num_replicates + 1))


@pytest.mark.parametrize("window_size", [44, 88, 176])
def test_from_dir(
    specific_config,
    labelset_notmat,
    tmp_path,
    default_model,
    device,
    previous_run_path_factory,
    window_size,
):
    root_results_dir = tmp_path.joinpath("tmp_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "LEARNCURVE", "option": "device", "value": device},
        {
            "section": "LEARNCURVE",
            "option": "previous_run_path",
            "value": str(previous_run_path_factory(default_model)),
        },
        {"section": "DATALOADER", "option": "window_size", "value": window_size},
    ]
    toml_path = specific_config(
        config_type="learncurve",
        model=default_model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)

    csv_path = cfg.learncurve.csv_path
    dataset_df = pd.read_csv(csv_path)
    timebin_dur = vak.io.dataframe.validate_and_get_timebin_dur(dataset_df)

    labelset_notmat = vak.converters.labelset_to_set(labelset_notmat)
    has_unlabeled = vak.csv.has_unlabeled(csv_path, labelset_notmat, TIMEBINS_KEY)
    if has_unlabeled:
        map_unlabeled = True
    else:
        map_unlabeled = False
    labelmap = vak.labels.to_map(labelset_notmat, map_unlabeled=map_unlabeled)

    results_path = tmp_path.joinpath("test_from_dir")
    results_path.mkdir()

    previous_run_path = previous_run_path_factory(default_model)
    previous_run_path = vak.converters.expanded_user_path(previous_run_path)

    train_dur_csv_paths = vak.core.learncurve.train_dur_csv_paths.from_dir(
        previous_run_path,
        cfg.learncurve.train_set_durs,
        timebin_dur,
        cfg.learncurve.num_replicates,
        results_path,
        window_size,
        SPECT_KEY,
        TIMEBINS_KEY,
        labelmap,
    )
    assert isinstance(train_dur_csv_paths, dict)
    assert sorted(train_dur_csv_paths.keys()) == sorted(cfg.learncurve.train_set_durs)
    for train_set_dur, csv_list in train_dur_csv_paths.items():
        replicate_nums = [int(csv.parent.name.split("_")[-1]) for csv in csv_list]
        assert replicate_nums == list(range(1, cfg.learncurve.num_replicates + 1))


@pytest.mark.parametrize("window_size", [44, 88, 176])
def test_from_df(
    specific_config, labelset_notmat, default_model, device, tmp_path, window_size
):
    root_results_dir = tmp_path.joinpath("tmp_root_results_dir")
    root_results_dir.mkdir()

    options_to_change = [
        {
            "section": "LEARNCURVE",
            "option": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"section": "LEARNCURVE", "option": "device", "value": device},
        {"section": "DATALOADER", "option": "window_size", "value": window_size},
    ]
    toml_path = specific_config(
        config_type="learncurve",
        model=default_model,
        audio_format="cbin",
        annot_format="notmat",
        options_to_change=options_to_change,
    )
    cfg = vak.config.parse.from_toml_path(toml_path)

    csv_path = cfg.learncurve.csv_path
    dataset_df = pd.read_csv(csv_path)
    timebin_dur = vak.io.dataframe.validate_and_get_timebin_dur(dataset_df)

    labelset_notmat = vak.converters.labelset_to_set(labelset_notmat)
    has_unlabeled = vak.csv.has_unlabeled(csv_path, labelset_notmat, TIMEBINS_KEY)
    if has_unlabeled:
        map_unlabeled = True
    else:
        map_unlabeled = False
    labelmap = vak.labels.to_map(labelset_notmat, map_unlabeled=map_unlabeled)

    results_path = tmp_path.joinpath("test_from_df")
    results_path.mkdir()

    train_dur_csv_paths = vak.core.learncurve.train_dur_csv_paths.from_df(
        dataset_df,
        csv_path,
        cfg.learncurve.train_set_durs,
        timebin_dur,
        cfg.learncurve.num_replicates,
        results_path,
        labelmap,
        window_size,
        SPECT_KEY,
        TIMEBINS_KEY,
        labelmap,
    )



from datetime import datetime
import os
import time

import pytest

import vak.constants
import vak.paths


def results_path_matches_expected(
    results_path, root_results_dir, before_timestamp, after_timestamp
):
    assert results_path.parent == root_results_dir
    assert results_path.name.startswith(vak.constants.RESULTS_DIR_PREFIX)
    results_dirname_parts = results_path.name.split(sep="_")
    assert len(results_dirname_parts) == 3  # 'results', 'date', 'time'
    timenow_str = results_dirname_parts[1] + "_" + results_dirname_parts[2]
    assert len(timenow_str) == len(vak.constants.STRFTIME_TIMESTAMP)
    timenow_str_as_timestamp = time.mktime(
        datetime.strptime(timenow_str, vak.constants.STRFTIME_TIMESTAMP).timetuple()
    )
    assert before_timestamp <= timenow_str_as_timestamp <= after_timestamp
    return True


def test_generate_results_dir_name_as_path(tmp_path):
    before_timestamp = time.mktime(datetime.now().timetuple())
    results_path = vak.paths.generate_results_dir_name_as_path(
        root_results_dir=tmp_path
    )
    after_timestamp = time.mktime(datetime.now().timetuple())

    assert results_path_matches_expected(
        results_path,
        root_results_dir=tmp_path,
        before_timestamp=before_timestamp,
        after_timestamp=after_timestamp,
    )


def test_generate_results_dir_name_as_path_no_root_results_dir(tmp_path):
    before_timestamp = time.mktime(datetime.now().timetuple())
    home = os.getcwd()
    os.chdir(tmp_path)
    results_path = vak.paths.generate_results_dir_name_as_path(root_results_dir=None)
    after_timestamp = time.mktime(datetime.now().timetuple())

    results_path = results_path.resolve()
    assert results_path_matches_expected(
        results_path,
        root_results_dir=tmp_path,
        before_timestamp=before_timestamp,
        after_timestamp=after_timestamp,
    )
    os.chdir(home)


def test_generate_results_dir_name_as_path_nonexistent_root_raises():
    with pytest.raises(NotADirectoryError):
        vak.paths.generate_results_dir_name_as_path(
            root_results_dir="/obviously/not/an/existent/directory"
        )



from datetime import datetime

import crowsetta
import numpy as np

from . import audio, spect
from .. import annotation
from ..converters import expanded_user_path, labelset_to_set
from ..logging import log_or_print


def from_files(
    data_dir,
    annot_format=None,
    labelset=None,
    annot_file=None,
    audio_format=None,
    spect_format=None,
    spect_params=None,
    spect_output_dir=None,
    logger=None,
):
    """create a pandas DataFrame representing a dataset for machine learning
    from a set of files in a directory

    Prepares dataset of vocalizations from a directory of audio or spectrogram files,
    and (optionally) annotation for those files. The dataset is returned as a pandas DataFrame.

    Datasets are used to train neural networks, predicting annotations for
    the dataset itself using a trained neural network, etc.

    If dataset is created from audio files, then array files containing spectrograms
    will be generated from the audio files and saved in ``spect_output_dir``
    with the extension ``.spect.npz``. The ``spect_output_dir`` defaults to ``data_dir``
    if is not specified.

    Parameters
    ----------
    data_dir : str
        path to directory with audio or spectrogram files from which to make dataset
    annot_format : str
        format of annotations. Any format that can be used with the
        crowsetta library is valid. Default is None.
    labelset : str, list, set
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, then files will be skipped where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using ``vak.converters.labelset_to_set``.
        See help for that function for details on how to specify labelset.
    load_spects : bool
        if True, load spectrograms. If False, return a VocalDataset without spectograms loaded.
        Default is True. Set to False when you want to create a VocalDataset for use
        later, but don't want to load all the spectrograms into memory yet.
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}.
    spect_format : str
        format of array files containing spectrograms as 2-d matrices.
        One of {'mat', 'npz'}.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    spect_params : dict, vak.config.spect.SpectParamsConfig.
        Parameters for creating spectrograms.
        Default is None (implying that spectrograms are already made).
    spect_output_dir : str
        path to location where spectrogram files should be saved.
        Default is None, in which case it defaults to ``data_dir``.
        A new directory will be created in ``spect_output_dir`` with
        the name 'spectrograms_generated_{time stamp}'.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    vak_df : pandas.DataFrame
        the dataset prepared from the directory specified
    """
    # ---- pre-conditions ----------------------------------------------------------------------------------------------
    if labelset is not None:
        labelset = labelset_to_set(labelset)

    if audio_format is None and spect_format is None:
        raise ValueError("Must specify either audio_format or spect_format")

    if audio_format and spect_format:
        raise ValueError(
            "Cannot specify both audio_format and spect_format, "
            "unclear whether to create spectrograms from audio files or "
            "use already-generated spectrograms from array files"
        )

    data_dir = expanded_user_path(data_dir)
    if not data_dir.is_dir():
        raise NotADirectoryError(f"data_dir not found: {data_dir}")

    if spect_output_dir:
        spect_output_dir = expanded_user_path(spect_output_dir)
        if not spect_output_dir.is_dir():
            raise NotADirectoryError(f"spect_output_dir not found: {spect_output_dir}")
    else:
        spect_output_dir = data_dir

    if annot_format is not None:
        if annot_file is None:
            annot_files = annotation.files_from_dir(
                annot_dir=data_dir, annot_format=annot_format
            )
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = scribe.from_file(annot_files)
        else:
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = scribe.from_file(annot_file)
        if isinstance(annot_list, crowsetta.Annotation):
            # if e.g. only one annotated audio file in directory, wrap in a list to make iterable
            # fixes https://github.com/NickleDave/vak/issues/467
            annot_list = [annot_list]
    else:  # if annot_format not specified
        annot_list = None

    # ------ if making dataset from audio files, need to make into array files first! ----------------------------------
    if audio_format:
        log_or_print(
            f"making array files containing spectrograms from audio files in: {data_dir}",
            logger=logger,
            level="info",
        )
        audio_files = audio.files_from_dir(data_dir, audio_format)

        timenow = datetime.now().strftime("%y%m%d_%H%M%S")
        spect_dirname = f"spectrograms_generated_{timenow}"
        spect_output_dir = spect_output_dir.joinpath(spect_dirname)
        spect_output_dir.mkdir()

        spect_files = audio.to_spect(
            audio_format=audio_format,
            spect_params=spect_params,
            output_dir=spect_output_dir,
            audio_files=audio_files,
            annot_list=annot_list,
            labelset=labelset,
        )
        spect_format = "npz"
    else:  # if audio format is None
        spect_files = None

    to_dataframe_kwargs = {
        "spect_format": spect_format,
        "labelset": labelset,
        "annot_list": annot_list,
        "annot_format": annot_format,
    }

    if spect_files:  # because we just made them, and put them in spect_output_dir
        to_dataframe_kwargs["spect_files"] = spect_files
        log_or_print(
            f"creating dataset from spectrogram files in: {spect_output_dir}",
            logger=logger,
            level="info",
        )
    else:
        to_dataframe_kwargs["spect_dir"] = data_dir
        log_or_print(
            f"creating dataset from spectrogram files in: {data_dir}",
            logger=logger,
            level="info",
        )

    vak_df = spect.to_dataframe(**to_dataframe_kwargs, logger=logger)
    return vak_df


def add_split_col(df, split):
    """add a 'split' column to a pandas DataFrame.
    Useful for assigning an entire dataset to the same "split",
    e.g. 'train' or 'predict'.
    All rows in the 'split' column will have the value specified.

    Parameters
    ----------
    df : pandas.DataFrame
        that represents a dataset of vocalizations
    split : str
        string that will be assigned to every row in the added "split" column.
        One of {'train', 'val', 'test', 'predict'}.
    """
    if split not in {"train", "val", "test", "predict"}:
        raise ValueError(
            f"value for split should be one of {{'train', 'val', 'test', 'predict'}}, but was {split}"
        )
    split_col = np.asarray([split for _ in range(len(df))], dtype="object")
    df["split"] = split_col
    return df


def validate_and_get_timebin_dur(df, expected_timebin_dur=None):
    """validate timebin duration for a dataset represented by a pandas DataFrame

    checks that there is a single, unique value for the time bin duration of all
    spectrograms in the dataset, and if so, returns it

    Parameters
    ----------
    df : pandas.Dataframe
        created by dataframe.from_files or spect.to_dataframe
    expected_timebin_dur : float

    Returns
    -------
    timebin_dur : float
        duration of time bins for all spectrograms in the dataset
    """
    timebin_dur = df["timebin_dur"].unique()
    if len(timebin_dur) > 1:
        raise ValueError(
            f"found more than one time bin duration in dataset: {timebin_dur}"
        )
    elif len(timebin_dur) == 1:
        timebin_dur = timebin_dur.item()

    if expected_timebin_dur:
        if timebin_dur != expected_timebin_dur:
            raise ValueError(
                "timebin duration from dataset, {}, did not match expected timebin duration"
            )

    return timebin_dur


def split_dur(df, split):
    """get duration of a split in the dataset"""
    return df[df["split"] == split]["duration"].sum()



import pytest
import toml

import vak.config.validators


def test_are_sections_valid(invalid_section_config_path):
    """test that invalid section name raises a ValueError"""
    with invalid_section_config_path.open("r") as fp:
        config_toml = toml.load(fp)
    with pytest.raises(ValueError):
        vak.config.validators.are_sections_valid(
            config_toml, invalid_section_config_path
        )


def test_are_options_valid(invalid_option_config_path):
    """test that section with an invalid option name raises a ValueError"""
    section_with_invalid_option = "PREP"
    with invalid_option_config_path.open("r") as fp:
        config_toml = toml.load(fp)
    with pytest.raises(ValueError):
        vak.config.validators.are_options_valid(
            config_toml, section_with_invalid_option, invalid_option_config_path
        )



import numpy as np


def levenshtein(source, target):
    """Levenshtein distance: number of deletions, insertions,
    or substitutions required to convert source string
    into target string.

    Parameters
    ----------
    source, target : str

    Returns
    -------
    distance : int
        number of deletions, insertions, or substitutions
        required to convert source into target.

    adapted from https://github.com/toastdriven/pylev/blob/master/pylev.py
    to fix issues with the Numpy implementation in
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
    """
    if source == target:
        return 0

    # We call tuple() to force strings to be used as sequences
    # ('c', 'a', 't', 's') - numpy uses them as values by default.
    source = np.array(tuple(source))
    target = np.array(tuple(target))

    len_source = source.size
    len_target = target.size

    if len_source == 0:
        return len_target
    if len_target == 0:
        return len_source

    if len_source > len_target:
        source, target = target, source
        len_source, len_target = len_target, len_source

    # We use a dynamic programming algorithm, but with the
    # added optimization that we only need the last two rows
    # of the matrix.
    d0 = np.arange(len_target + 1)
    d1 = np.arange(len_target + 1)
    for i in range(len_source):
        d1[0] = i + 1
        for j in range(len_target):
            cost = d0[j]

            if source[i] != target[j]:
                cost += 1  # substitution

                x_cost = d1[j] + 1  # insertion
                if x_cost < cost:
                    cost = x_cost

                y_cost = d0[j + 1] + 1
                if y_cost < cost:
                    cost = y_cost

            d1[j + 1] = cost

        d0, d1 = d1, d0

    return d0[-1]


def segment_error_rate(y_pred, y_true):
    """Levenshtein edit distance normalized by length of true sequence.
    Also known as word error distance; here applied to other vocalizations
    in addition to speech.

    Parameters
    ----------
    y_pred : str
        predicted labels for a series of songbird syllables
    y_true : str
        ground truth labels for a series of songbird syllables

    Returns
    -------
    Levenshtein distance / len(y_true)
    """
    if type(y_true) != str or type(y_pred) != str:
        raise TypeError("Both `y_true` and `y_pred` must be of type `str")

    # handle divide by zero edge cases
    if len(y_true) == 0 and len(y_pred) == 0:
        return 0.
    elif len(y_true) == 0 and len(y_pred) != 0:
        raise ValueError(
            f'segment error rate is undefined when length of y_true is zero'
        )

    return levenshtein(y_pred, y_true) / len(y_true)
